""" Module for Amino Acid Record """
from __future__ import annotations
from typing import Iterable, List, Set, Tuple, Union, Pattern
import copy
import re
import regex
from Bio.SeqRecord import SeqRecord
from Bio import SeqUtils
from Bio.Seq import Seq
from moPepGen.SeqFeature import MatchedLocation, FeatureLocation
from moPepGen.aa.expasy_rules import EXPASY_RULES, EXPASY_RULES2, EXPASY_RULES_WINGS_SIZE


class AminoAcidSeqRecord(SeqRecord):
    #pylint: disable=W0223
    """ A AminoAcidSeqRecord holds a protein or peptide sequence
    """
    def __init__(self, seq:Seq, _id:str="<unknown id>",
            name:str="<unknown name>", description:str="<unknown description>",
            gene_id:str=None, transcript_id:str=None, protein_id:str=None,
            gene_name:str=None, **kwargs):
        self.id = None
        self.name = None
        super().__init__(seq, id=_id, name=name, description=description,
            **kwargs)
        self.gene_id = gene_id
        self.transcript_id = transcript_id
        self.protein_id = protein_id
        self.gene_name = gene_name

    def __getitem__(self, index) -> AminoAcidSeqRecord:
        """"""
        new_seq = self.seq.__getitem__(index)
        new_one = self.__class__(seq=new_seq, _id=self.id, name=self.name,
            description=self.description, gene_id=self.gene_id,
            transcript_id=self.transcript_id, protein_id=self.protein_id)
        return new_one

    def __add__(self, other:AminoAcidSeqRecord) -> AminoAcidSeqRecord:
        """"""
        new_one = super().__add__(other)
        new_one.__class__ = self.__class__
        new_one.gene_id = self.gene_id
        new_one.transcript_id = self.transcript_id
        new_one.protein_id = self.protein_id
        return new_one

    def __hash__(self):
        """ hash """
        return hash(str(self.seq))

    def __eq__(self, other:AminoAcidSeqRecord):
        """ Equal to. It is implemented in this way, because the get_equivalent
        calls this __eq__ instead of the other for reason that I don't
        understand. """
        result = (self.seq == other.seq)
        if result and hasattr(other, 'match'):
            other.match = self
        return result

    def __ne__(self, other:AminoAcidSeqRecord) -> bool:
        """ not equal to """
        return not self == other

    def infer_ids(self, style:str=None) -> str:
        """
        Args:
            source (str): The style of the fasta file header. Either 'GENCODE'
                or 'ENSEMBL'. If None, it will try both. Defaults to None.

        Returns:
            The style ifered.
        """
        if not style:
            try:
                self.infer_ids_gencode()
                return 'GENCODE'
            except ValueError:
                pass

            try:
                self.infer_ids_ensembl()
                return 'ENSEMBL'
            except ValueError as e:
                raise ValueError(
                    'Failed to infer gene ID, transcript ID, and protein ID '
                    r'using both ENSEMBL and GENCODE\'s format.'
                ) from e
        elif style == 'GENCODE':
            self.infer_ids_gencode()
        elif style == 'ENSEMBL':
            self.infer_ids_ensembl()
        else:
            raise ValueError(f'style {style} is not supported')
        return style


    def infer_ids_gencode(self):
        """ Infers the gene, transcript, and protein ID from description base
        on the GENCODE format
        """
        p = r'^(?P<protein_id>[A-Z0-9.:_-]+)\|' +\
            r'(?P<transcript_id>[A-Z0-9.:_-]+)\|(?P<gene_id>[A-Z0-9._]+)\|.+'

        match = re.search(p, self.description)
        if not match:
            raise ValueError(
                f"The record doesn't seem to follow the GENCODE format: {self.description}"
            )

        protein_id = match.group('protein_id')
        transcript_id = match.group('transcript_id')
        gene_id = match.group('gene_id')

        self.id = protein_id
        self.name = protein_id
        self.gene_id = gene_id
        self.protein_id = protein_id
        self.transcript_id = transcript_id

    def infer_ids_ensembl(self):
        """ Infers the gene, transcript, and protein ID from description base
        on the ENSEMBL format
        """
        p = r'^(?P<protein_id>.+) pep (\bchromosome\b|\bsupercontig\b|\bscaffold\b):' +\
            r'(?P<genome_build>.+?):(?P<chromosome>.+?):(?P<position>\d+:\d+):(?P<strand>.+) ' +\
            r'gene:(?P<gene_id>.+) transcript:(?P<transcript_id>\S+) .+'

        match = re.search(p, self.description)
        if not match:
            raise ValueError(f"doesn't seem to follow the ENSEMBL format: {self.description}")

        protein_id = self.id
        gene_id = match.group('gene_id')
        transcript_id = match.group('transcript_id')

        self.gene_id = gene_id.split('.')[0]
        self.protein_id = protein_id.split('.')[0]
        self.transcript_id = transcript_id.split('.')[0]

    def iter_stop_sites(self) -> Iterable[int]:
        """ Create a generator for stop sites """
        for occ in re.finditer(r'\*', str(self.seq)):
            yield occ.start()

    def iter_enzymatic_cleave_sites(self, rule:str, exception:str=None,
            exception_sites:List[int]=None) -> Iterable[int]:
        """ Create a generator of the cleave sites """
        rule = EXPASY_RULES[rule]
        exception = EXPASY_RULES.get(exception, exception)
        if exception_sites is None:
            exception_sites = [] if exception is None else \
                [x.end() for x in re.finditer(exception, str(self.seq))]

        for it in re.finditer(rule, str(self.seq)):
            s = it.end()
            if s not in exception_sites:
                yield it.end()

    def iter_enzymatic_cleave_sites_with_range(self, rule:str, exception:str=None,
            exception_sites:List[int]=None
            ) -> Iterable[Tuple[int, Union[Tuple[int,int], None]]]:
        """ Create a generator of the cleave sites and also yields the full
        range of the pattern recognized during enzymatic cleavage.

        ## Examples:

        In this example, the cleavage site is 5 (after K), and the pattern range
        is (4,6) for KT.
        >>> seq = AminoAcidSeqRecord(Seq('TTTTKTTTT'))
        >>> list(seq.iter_enzymatic_cleave_sites_with_range('trypsin'))
        [(5,(4,6))]

        In this example, the cleavage site is 5 (after R) and the pattern range
        is (3,6) for MRP.
        >>> seq = AminoAcidSeqRecord(Seq('TTTMRPTTT'))
        >>> list(seq.iter_enzymatic_cleave_sites_with_range('trypsin'))
        [(5,(3,6))]
        """
        site_pattern = re.compile(EXPASY_RULES[rule])
        range_pattern = regex.compile(EXPASY_RULES2[rule])
        exception = EXPASY_RULES.get(exception, exception)
        seq = str(self.seq)
        if exception_sites is None:
            exception_sites = [] if exception is None else \
                [x.end() for x in re.finditer(exception, seq)]

        sites = [x.end() for x in site_pattern.finditer(seq)]
        ranges = [(x.start(), x.end()) for x in range_pattern.finditer(seq, overlapped=True)]

        if len(sites) != len(ranges):
            raise ValueError(
                f"Inconsistent cleavage sites found. seq={seq}, "
                f"sites={sites}, ranges={ranges}"
            )
        for s, r in zip(sites, ranges):
            if s not in exception_sites:
                yield s, r

    def iter_enzymatic_cleave_sites_with_range_local(self, rule:str, exception:str=None,
            exception_sites:List[int]=None
            ) -> Iterable[Tuple[int, Union[Tuple[int,int], None]]]:
        """ Iter enzymatic cleavage sites with pattern matching ranges using the
        local matching method. """
        rule_pattern = re.compile(EXPASY_RULES[rule])
        exception = EXPASY_RULES.get(exception, exception)
        wings_size = EXPASY_RULES_WINGS_SIZE[rule]
        seq = str(self.seq)
        seq_len = len(seq)
        if exception_sites is None:
            exception_sites = [] if exception is None else \
                [x.end() for x in re.finditer(exception, seq)]

        sites = [x.end() for x in rule_pattern.finditer(seq)]

        if wings_size[0] <= 0 and wings_size[1] <= 0:
            raise ValueError(
                "Invalid enzyme pattern with the size being 0 for both wings:" +
                f"{rule}: {EXPASY_RULES[rule]}"
            )

        for s in sites:
            if s in exception_sites:
                continue
            r = self.get_local_matched_range(
                seq=seq, site=s, p=rule_pattern, wings_size=wings_size, seq_len=seq_len
            )
            yield s, r

    @staticmethod
    def get_local_matched_range(seq:str, site:int, p:Pattern[str],
            wings_size:Tuple[int, int], seq_len:int=None):
        """ Get the local matched pattern range from a sequence
        with a given cleavage site. """
        if not seq_len:
            seq_len = len(seq)

        upper = max(site - wings_size[0], 0)
        lower = min(site + wings_size[1], len(seq))

        if wings_size[0] >= wings_size[1]:
            ucur = site - 1
            lcur = site
        else:
            ucur = site
            lcur = site + 1

        pattern_found: bool = False
        while True:
            if not upper <= ucur < lcur <= lower:
                break
            local_seg = seq[ucur:lcur]
            if p.search(local_seg):
                pattern_found = True
                break
            if (site - ucur) <= (lcur - site) or lcur == lower:
                ucur -= 1
            else:
                lcur += 1
        if not pattern_found:
            raise ValueError(f"Cannot extract matched pattern at position {site} from {seq}")
        return (ucur, lcur)

    def get_enzymatic_cleave_exception_sites(self, exception:str) -> Iterable[int]:
        """ Find all enzymatic cleavage exception sites """
        return [] if exception is None else \
            [x.end() for x in re.finditer(exception, str(self.seq))]

    def find_first_cleave_or_stop_site(self, rule:str, exception:str=None,
            exception_sites:List[int]=None) -> Iterable[int]:
        """ Create a generator for cleave or stop sites """
        it_cleavage = self.iter_enzymatic_cleave_sites(
            rule=rule, exception=exception, exception_sites=exception_sites
        )
        sites = []
        first_cleave_site = next(it_cleavage, None)
        if first_cleave_site is not None:
            sites.append(first_cleave_site)
        it_stop = self.iter_stop_sites()
        first_stop_site = next(it_stop, None)
        if first_stop_site is not None:
            if first_stop_site == 0:
                if len(self.seq) == 1:
                    return -1
                sites.append(first_stop_site + 1)
            else:
                sites.append(first_stop_site)
        if not sites:
            return -1
        return min(sites)

    def find_first_cleave_or_stop_site_with_range(self, rule:str, exception:str=None,
            exception_sites:List[int]=None
            ) -> Tuple[int, Union[Tuple[int,int], None]]:
        """ Create a generator for cleave or stop sites """
        it_cleavage = self.iter_enzymatic_cleave_sites_with_range(
            rule=rule, exception=exception, exception_sites=exception_sites
        )
        sites:List[Tuple[int, Union[Tuple[int,int], None]]] = []
        first_cleave_site, first_range = next(it_cleavage, (None, None))
        if first_cleave_site is not None:
            sites.append((first_cleave_site, first_range))
        it_stop = self.iter_stop_sites()
        first_stop_site = next(it_stop, None)
        if first_stop_site is not None:
            if first_stop_site == 0:
                if len(self.seq) == 1:
                    return -1, None
                sites.append((first_stop_site + 1, None))
            else:
                sites.append((first_stop_site, None))
        if not sites:
            return -1, None
        return min(sites, key=lambda x: x[0])

    def find_first_enzymatic_cleave_site(self, rule:str, exception:str=None,
            start:int=0) -> int:
        """ Find the first enzymatic cleave site """
        it = self[start:]\
            .iter_enzymatic_cleave_sites(rule=rule, exception=exception)
        try:
            return next(it) + start
        except StopIteration:
            return -1

    def find_all_enzymatic_cleave_sites(self, rule:str, exception:str=None,
            ) -> List[int]:
        """ Find all enzymatic cleave sites. """
        return list(self.iter_enzymatic_cleave_sites(rule=rule,
            exception=exception))

    def find_all_enzymatic_cleave_sites_with_ranges(self, rule:str, exception:str=None,
            ) -> List[Tuple[int, Tuple[int, int]]]:
        """ Find all enzymatic cleave sites. """
        return list(self.iter_enzymatic_cleave_sites_with_range(rule=rule,
            exception=exception))

    def find_all_start_sites(self) -> List[int]:
        """ Find all start positions """
        return [x.start() for x in re.finditer('M', str(self.seq))]

    def find_all_cleave_and_stop_sites(self, rule:str, exception:str=None,
            exception_sites:List[int]=None) -> List[int]:
        """ Find all enzymatic lceave sites and stop sites """
        cleavage_sites = list(self.iter_enzymatic_cleave_sites(rule=rule,
            exception=exception, exception_sites=exception_sites))
        stop_sites_start = list(self.iter_stop_sites())
        stop_sites_end = [i + 1 for i in stop_sites_start if i < len(self.seq) - 1]
        stop_sites_start = [i for i in stop_sites_start if i > 0]
        sites = cleavage_sites + stop_sites_start + stop_sites_end
        sites = list(set(sites))
        sites.sort()
        return sites

    def find_all_cleave_and_stop_sites_with_range(self, rule:str,
            exception:str=None, exception_sites:List[int]=None
            ) -> List[Tuple[int,Union[Tuple[int,int], None]]]:
        """ Find all enzymatic cleavage sites with ranges and stop sites """
        sites = list(
            self.iter_enzymatic_cleave_sites_with_range(
                rule=rule, exception=exception,
                exception_sites=exception_sites
            )
        )
        sites_mapper = dict(sites)
        stop_sites_start = list(self.iter_stop_sites())
        stop_sites_end = [i + 1 for i in stop_sites_start if i < len(self.seq) - 1]
        stop_sites_start = [i for i in stop_sites_start if i > 0]

        for s in stop_sites_start:
            sites_mapper[s] = (s, s+1)
        for s in stop_sites_end:
            sites_mapper[s] = (s-1, s)

        sites = list(sites_mapper.items())
        sites.sort(key=lambda x: x[0])
        return sites

    def enzymatic_cleave(self, rule:str, exception:str=None,
            miscleavage:int=2, min_mw:float=500.0, min_length:int=7,
            max_length:int=25, cds_start_nf:bool=False
            )->Set[AminoAcidSeqRecord]:
        """ Performs enzymatic cleave """
        peptides = []
        sites = [0]
        sites += self.find_all_enzymatic_cleave_sites(
            rule=rule, exception=exception)
        sites.append(len(self))
        start = 0

        def update_peptides(peptide):
            if 'X' in peptide.seq:
                return
            mol_wt = SeqUtils.molecular_weight(peptide.seq, 'protein')
            weight_flag = mol_wt > min_mw
            length_flag = len(peptide.seq) >= min_length \
                and len(peptide.seq) <= max_length
            if weight_flag and length_flag:
                peptides.append(peptide)

        while start < len(sites) - 1:
            end = start + 1
            while end - start - 1 <= miscleavage and end < len(sites):
                peptide = self[sites[start]:sites[end]]
                if start == 0 and not cds_start_nf and peptide.seq.startswith('M'):
                    update_peptides(peptide[1:])
                update_peptides(peptide)
                end += 1
            start += 1
        return peptides


class AminoAcidSeqRecordWithCoordinates(AminoAcidSeqRecord):
    """ Amino acid sequence record with coordinates """
    def __init__(self, seq:Seq, *args,
            locations:List[MatchedLocation]=None, orf:FeatureLocation=None, **kwargs ):
        """ Constract a DNASeqRecordWithCoordinates object.

        Args:
            seq (Seq): The amino acid sequence.
            locations (List[MatchedLocation]): The locations where this
                sequence is aligned tp.
            orf (FeatureLocation): The open reading frame start and end.
        """
        super().__init__(seq=seq, *args, **kwargs)
        self.locations = locations or []
        # query index
        self.orf = orf

    def __add__(self, other:AminoAcidSeqRecordWithCoordinates
            ) -> AminoAcidSeqRecordWithCoordinates:
        """ add operation """
        new = super().__add__(other)
        new.__class__ = AminoAcidSeqRecordWithCoordinates
        left_locs = copy.copy(self.locations)
        right_locs = copy.copy(other.locations)
        if left_locs and right_locs:
            lhs = self.locations[-1]
            rhs = other.locations[0].shift(len(self))
            if lhs.query.end == rhs.query.start \
                    and lhs.ref.end == rhs.ref.start \
                    and lhs.get_ref_dna_end() == rhs.get_ref_dna_start():
                query = FeatureLocation(
                    start=lhs.query.start, end=rhs.query.end,
                    reading_frame_index=lhs.query.reading_frame_index,
                    start_offset=lhs.query.start_offset,
                    end_offset=rhs.query.end_offset
                )
                ref = FeatureLocation(
                    start=lhs.ref.start, end=rhs.ref.end, seqname=lhs.ref.seqname,
                    reading_frame_index=lhs.ref.reading_frame_index,
                    start_offset=lhs.ref.start_offset,
                    end_offset=rhs.ref.end_offset
                )
                new_loc = MatchedLocation(query=query, ref=ref)
                right_locs.pop(0)
                left_locs[-1] = new_loc

        new.locations = left_locs + [loc.shift(len(self)) for loc in right_locs]
        new.orf = self.orf
        return new

    def __getitem__(self, index)->AminoAcidSeqRecordWithCoordinates:
        """ get item """
        if isinstance(index, int):
            return super().__getitem__(index)
        start, stop, _ = index.indices(len(self))

        locations = []
        if start != stop:
            for location in self.locations:
                lhs = location.query.start
                rhs = location.query.end
                if rhs <= start:
                    continue
                if lhs >= stop:
                    break
                if lhs <= start:
                    if rhs <= stop:
                        location = location[start-lhs:]
                    else:
                        location = location[start-lhs:stop-lhs]
                else:
                    if rhs <= stop:
                        location = location.shift(-start)
                    else:
                        new_location = location[:stop-lhs]
                        new_location = new_location.shift(lhs-start)
                        location = new_location

                locations.append(location)

        return self.__class__(
            seq=self.seq[index],
            locations=locations,
            orf = self.orf,
            _id=self.id,
            name=self.name,
            description=self.description
        )

    def __hash__(self):
        """hash"""
        return hash(self.seq)

    def get_query_index(self, ref_index:int, seqname:str, reading_frame:int=None) -> int:
        """ Returns the query index wiht a given reference index """
        for location in self.locations:
            if location.ref.seqname != seqname:
                continue
            if reading_frame is not None and location.query.reading_frame_index != reading_frame:
                continue
            if ref_index in location.ref:
                return location.query.start + ref_index - location.ref.start
        return -1



""" Module for rMATS parsers """
from typing import Iterable
from .RMATSRecord import RMATSRecord
from .SERecord import SERecord
from .A5SSRecord import A5SSRecord
from .A3SSRecord import A3SSRecord
from .MXERecord import MXERecord
from .RIRecord import RIRecord

def parse(path:str, event_type:str) -> Iterable[RMATSRecord]:
    """ parse """
    with open(path, 'rt') as handle:
        # first line is header
        next(handle, None)
        for line in handle:
            if event_type == 'SE':
                yield SERecord.readline(line)
            if event_type == 'A5SS':
                yield A5SSRecord.readline(line)
            if event_type == 'A3SS':
                yield A3SSRecord.readline(line)
            if event_type == 'MXE':
                yield MXERecord.readline(line)
            if event_type == 'RI':
                yield RIRecord.readline(line)



""" moPepGen index """
from __future__ import annotations
from typing import TYPE_CHECKING, Set, List
import os
import shutil
from pathlib import Path
import gzip
import json
import pickle
from moPepGen import gtf, logger, err
from moPepGen.version import MetaVersion
from moPepGen.params import CleavageParams


if TYPE_CHECKING:
    from moPepGen.dna import DNASeqDict
    from moPepGen.gtf import GenomicAnnotationOnDisk
    from moPepGen.aa import AminoAcidSeqDict
    from moPepGen.gtf.GTFPointer import TranscriptPointer, GenePointer

class CanonicalPoolMetadata:
    """ Canonical peptide pool metadata """
    def __init__(self, filename:str, index:int, cleavage_params:CleavageParams):
        """ constructor """
        self.filename = filename
        self.index = index
        self.cleavage_params = cleavage_params

    def jsonfy(self):
        """ jsonfy """
        return {
            'filename': self.filename,
            'index': self.index,
            'cleavage_params': self.cleavage_params.jsonfy()
        }

class IndexMetadata:
    """ Index metadata """
    def __init__(self, version:MetaVersion, canonical_pools:List[CanonicalPoolMetadata],
            source:str):
        """ constructor """
        self.version = version
        self.canonical_pools = canonical_pools
        self.source = source

    def jsonfy(self):
        """ jsonfy """
        return {
            'version': self.version.jsonfy(),
            'canonical_pools': [it.jsonfy() for it in self.canonical_pools],
            'source': self.source
        }

    def register_canonical_pool(self, cleavage_params:CleavageParams):
        """ Register for new canonical pool """
        if self.get_canonical_pool(cleavage_params):
            raise ValueError(
                "Canonical peptide pool already exists with the parameters."
            )
        index = max(it.index for it in self.canonical_pools) + 1 if self.canonical_pools else 1
        pool = CanonicalPoolMetadata(
            filename = f"canonical_peptides_{index:03}.pkl",
            index=index,
            cleavage_params=cleavage_params
        )
        self.canonical_pools.append(pool)
        return pool

    def get_canonical_pool(self, cleavage_params:CleavageParams) -> CanonicalPoolMetadata:
        """ Get canonical peptide pool metadata that match with the given
        cleavage parameters. """
        this = cleavage_params.jsonfy(graph_params=False)
        for pool in self.canonical_pools:
            that = pool.cleavage_params.jsonfy(graph_params=False)
            if this == that:
                return pool
        return None

class IndexDir:
    """ moPepGen index directory """
    def __init__(self, path:Path):
        """ constructor """
        self.path = path
        self.genome_file = self.path/'genome.pkl'
        self.annotation_file = self.path/'annotation.gtf'
        self.proteome_file = self.path/'proteome.pkl'
        self.coding_tx_file = self.path/'coding_transcripts.pkl'
        self.metadata_file = self.path/'metadata.json'
        if self.metadata_file.exists():
            self.metadata = self.load_metadata()
        else:
            self.init_metadata()

    def init_metadata(self):
        """ Initialize metadata """
        self.metadata = IndexMetadata(
            version=MetaVersion(),
            canonical_pools=[],
            source=None
        )

    def load_metadata(self) -> IndexMetadata:
        """ Get metadata of the moPepGen index """
        with open(self.metadata_file, 'rt') as handle:
            data = json.load(handle)
            version = MetaVersion(**data['version'])
            canonical_pools = []
            for it in data['canonical_pools']:
                cleavage_params = CleavageParams(**it['cleavage_params'])
                pool = CanonicalPoolMetadata(
                    filename=it['filename'],
                    index=it['index'],
                    cleavage_params=cleavage_params
                )
                canonical_pools.append(pool)
            source = data['source']
        return IndexMetadata(
            version=version,
            canonical_pools=canonical_pools,
            source=source
        )

    def validate_metadata(self) -> bool:
        """ Valid the index metadata """
        cur_version = MetaVersion()
        if not cur_version.is_valid(self.metadata.version):
            raise err.InvalidIndexError(cur_version, self.metadata.version)
        return True

    def save_metadata(self):
        """ Save metadata to index directory """
        data = self.metadata.jsonfy()
        with open(self.metadata_file, 'wt') as handle:
            json.dump(data, handle, indent=2)

    def load_genome(self) -> DNASeqDict:
        """ Load genome from index directory """
        with open(self.genome_file, 'rb') as handle:
            return pickle.load(handle)

    def save_genome(self, genome:DNASeqDict):
        """ Save genome to index directory """
        with open(self.genome_file, 'wb') as handle:
            pickle.dump(genome, handle)

    def load_annotation(self) -> GenomicAnnotationOnDisk:
        """ Load genomic annotation from index directory """
        anno = gtf.GenomicAnnotationOnDisk()
        anno.init_handle(self.annotation_file)
        anno.load_index(self.annotation_file, source=self.metadata.source)
        return anno

    def create_gtf_copy(self, file:Path, symlink:bool=True):
        """ Create copy of GTF """
        if file.suffix.lower() == '.gz':
            if symlink:
                symlink = False
                logger(
                    "--gtf-symlink was suppressed because compressed GTF file was received. "
                )
        elif file.suffix.lower() != '.gtf':
            raise ValueError(f"Cannot handle gtf file {file}")

        if symlink:
            os.symlink(file.absolute(), self.annotation_file)
        elif file.suffix.lower() == '.gtf':
            shutil.copy2(file, self.annotation_file)
        else:
            with gzip.open(file, 'rt') as ihandle, open(self.annotation_file, 'wt') as ohandle:
                for line in ihandle:
                    ohandle.write(line)

    def save_annotation(self, file:Path, source:str=None, proteome:AminoAcidSeqDict=None,
            invalid_protein_as_noncoding:bool=True, symlink:bool=True) -> GenomicAnnotationOnDisk:
        """ Save genomic annotation to index directory """
        self.create_gtf_copy(file, symlink=symlink)
        anno = gtf.GenomicAnnotationOnDisk()
        anno.generate_index(self.annotation_file, source)

        if proteome:
            anno.check_protein_coding(proteome, invalid_protein_as_noncoding)

        gene_idx_file, tx_idx_file = anno.get_index_files(self.annotation_file)

        with open(gene_idx_file, 'wt') as handle:
            for gene in anno.genes.keys():
                gene_pointer:GenePointer = anno.genes.get_pointer(gene)
                handle.write(gene_pointer.to_line() + '\n')

        with open(tx_idx_file, 'wt') as handle:
            for tx in anno.transcripts.keys():
                tx_pointer:TranscriptPointer = anno.transcripts.get_pointer(tx)
                handle.write(tx_pointer.to_line() + '\n')
        return anno

    def load_proteome(self) -> AminoAcidSeqDict:
        """ Load proteome from index directory """
        with open(self.proteome_file, 'rb') as handle:
            return pickle.load(handle)

    def save_proteome(self, proteome:AminoAcidSeqDict):
        """ Save proteome to index directory """
        with open(self.proteome_file, 'wb') as handle:
            pickle.dump(proteome, handle)

    def load_canonical_peptides(self, cleavage_params:CleavageParams) -> Set[str]:
        """ Load canonical peptides from index directory """
        pool_data = self.metadata.get_canonical_pool(cleavage_params)
        if not pool_data:
            raise ValueError(
                'No canonical peptide pool match with the cleavage parameters'
            )
        with open(self.path/pool_data.filename, 'rb') as handle:
            return pickle.load(handle)

    def save_canonical_peptides(self, seqs:Set[str], cleavage_params:CleavageParams,
            override:bool=False):
        """ Save canonical peptides to index directory """
        pool_metadata = self.metadata.get_canonical_pool(cleavage_params)
        if not pool_metadata or not override:
            pool_metadata = self.metadata.register_canonical_pool(cleavage_params)
        with open(self.path/pool_metadata.filename, 'wb') as handle:
            pickle.dump(seqs, handle)

    def wipe_canonical_peptides(self):
        """ Remove all canonical peptide pools """
        for pool in self.metadata.canonical_pools:
            os.remove(self.path/pool.filename)
        self.metadata.canonical_pools = []

    def load_coding_tx(self) -> Set[str]:
        """ Load coding transcripts from index directory """
        with open(self.coding_tx_file, 'rb') as handle:
            return pickle.load(handle)

    def save_coding_tx(self, coding_tx:Set[str]):
        """ Save coding transcripts to index directory """
        with open(self.coding_tx_file, 'wb') as handle:
            pickle.dump(coding_tx, handle)



""" Module for generic variant record """
from __future__ import annotations
import copy
from typing import TYPE_CHECKING, Dict, Iterable, List
from moPepGen import constant, ERROR_NO_TX_AVAILABLE, \
    ERROR_VARIANT_NOT_IN_GENE_COORDINATE, ERROR_INDEX_IN_INTRON, \
        ERROR_REF_LENGTH_NOT_MATCH_WITH_LOCATION
from moPepGen.SeqFeature import FeatureLocation


_VARIANT_TYPES = ['SNV', 'INDEL', 'MNV', 'Fusion', 'RNAEditingSite',
    'Insertion', 'Deletion', 'Substitution', 'circRNA', 'SECT', 'W2F']

# To avoid circular import
if TYPE_CHECKING:
    from moPepGen.gtf import GenomicAnnotation
    from moPepGen.dna import DNASeqDict, DNASeqRecord, DNASeqRecordWithCoordinates


def create_variant_sect(anno:GenomicAnnotation, tx_id:str, pos:int) -> VariantRecord:
    """ Create a VariantRecord for Selenocysteine Termination. """
    gene_id = anno.transcripts[tx_id].gene_id
    start_tx = pos
    end_tx = pos + 2
    start_genome = anno.coordinate_transcript_to_genomic(start_tx, tx_id)
    end_genome = anno.coordinate_transcript_to_genomic(end_tx, tx_id)
    start_gene = anno.coordinate_genomic_to_gene(start_genome, gene_id)
    end_gene = anno.coordinate_genomic_to_gene(end_genome, gene_id)
    end_gene += 1
    location = FeatureLocation(start_tx, end_tx + 1)
    ref = 'TGA'
    alt = '<SECT>'
    _id = f"SECT-{start_gene + 1}"
    attrs = {
        'TRANSCRIPT_ID': tx_id
    }
    return VariantRecord(location=location, ref=ref, alt=alt, _type='SECT',
        _id=_id, attrs=attrs)

def create_variant_w2f(tx_id:str, pos:int) -> VariantRecord:
    """ Create a W2F codon reassignment variant. """
    return VariantRecord(
        location=FeatureLocation(pos, pos + 1),
        ref='W',
        alt='F',
        _type='W2F',
        _id=f"W2F-{pos+1}",
        attrs={'TRANSCRIPT_ID': tx_id}
    )

def create_mnv_from_adjacent(variants:Iterable[VariantRecord]) -> VariantRecord:
    """ Create MNV from abdjacent variants. """
    var_ids = []
    for v in variants:
        if not var_ids:
            seqname = v.location.seqname
            if 'TRANSCRIPT_ID' in v.attrs:
                gene_id = seqname
                tx_id = v.attrs['TRANSCRIPT_ID']
            else:
                gene_id = v.attrs['GENE_ID']
                tx_id = seqname
            start = v.location.start
            ref = v.ref
            alt = v.alt
            var_ids.append(v.id)
        else:
            ref += v.ref
            alt += v.alt
            var_ids.append(v.id)
    end = variants[-1].location.end

    attrs = {
        'INDIVIDUAL_VARIANT_IDS': var_ids,
        'MERGED_MNV': True
    }
    if seqname == tx_id:
        attrs['GENE_ID'] = gene_id
    else:
        attrs['TRANSCRIPT_ID'] = tx_id

    return VariantRecord(
        location=FeatureLocation(start, end, seqname=seqname),
        ref=ref,
        alt=alt,
        _type='MNV',
        _id=f"MNV-{start}-{ref}-{alt}",
        attrs=attrs
    )

def find_mnvs_from_adjacent_variants(variants:List[VariantRecord],
        max_adjacent_as_mnv:int) -> List[VariantRecord]:
    """ Find MVNs from adjacent variants """
    mnvs = []
    compatible_type_map = {
        'SNV': 'SNV',
        'RNAEditingSite': 'SNV',
        'INDEL': 'INDEL'
    }
    for i,v_0 in enumerate(variants):
        if v_0.type not in compatible_type_map:
            continue
        type0 = compatible_type_map[v_0.type]
        adjacent_combs:Dict[int, List[int]] = {
            0: [[i]]
        }
        for k in range(1, max_adjacent_as_mnv):
            # if k not in adjacent_combs:
            #     break
            for comb in adjacent_combs[k - 1]:
                i_t = comb[-1]
                if i_t >= len(variants) - 1:
                    continue
                for j in range(i_t + 1, len(variants)):
                    v_j = variants[j]
                    if v_j.type not in compatible_type_map:
                        continue
                    if v_j.location.start < v_0.location.end:
                        continue
                    if v_j.location.start > v_0.location.end:
                        break
                    if compatible_type_map[v_j.type] == type0:
                        new_comb = comb + [j]
                        if k in adjacent_combs:
                            adjacent_combs[k].append(new_comb)
                        else:
                            adjacent_combs[k] = [new_comb]

        for k, combs in adjacent_combs.items():
            if k == 0:
                continue
            for comb in combs:
                mnv = create_mnv_from_adjacent([variants[x] for x in comb])
                mnvs.append(mnv)
    return mnvs

class VariantRecord():
    """ Defines the location, ref and alt of a genomic variant.

    A VCF-like definition is used here to represent a variant. For any type of
    variant, both the ref and alt are at least length of one.

    Examples:
        * SNP/SNV:
          The below example represents a SNP/SNV at location 101, that a A is
          substituted with T.
            >>> VariantRecord(
                    location=FeatureLocation(100, 101),
                    ref=DNASeqRecord(Seq('A')),
                    alt=DNASeqRecord(Seq('T'))
                )

        * insertion
          The example below represents a insertion between 115 and 116, and the
          inserted sequence is CTGAA
            >>> VariantRecord(
                    location=FeatureLocation(115, 116),
                    ref='G',
                    alt='GCTGAA'
                )

        * deletion
          The example below represents a deletion of CAC from position 80 to
          82.
            >>> VariantRecord(
                    location=FeatureLocation(79, 83),
                    ref='CCAC',
                    alt='C'
                )

    Attributes:
        location (FeatureLocation): The location of the variant.
        ref (str): Reference sequence
        alt (str): Altered sequence
    """
    def __init__(self, location:FeatureLocation, ref:str, alt:str, _type:str,
            _id:str, attrs:dict=None):
        """ Construct a VariantRecord object.

        Args:
            location (FeatureLocation): The location of the variant.
            ref (str): Reference sequence
            alt (str): Altered sequence
            _type (str): Variant type, must be one of 'SNV', 'INDEL', 'Fusion',
                'RNAEditingSite', 'AlternativeSplicing'
        """
        if _type not in ['Substitution', 'Deletion', 'circRNA'] and \
                len(location) != len(ref):
            raise ValueError(ERROR_REF_LENGTH_NOT_MATCH_WITH_LOCATION)
        if _type not in _VARIANT_TYPES:
            raise ValueError('Variant type not supported')
        self.location = location
        self.ref = ref
        self.alt = alt
        self.type = _type
        self.id = _id
        self.attrs = attrs if attrs else {}
        self.is_real_fusion = self.is_fusion()

    def __hash__(self):
        """ hash """
        donor_tx_id = self.attrs.get('DONOR_TRANSCRIPT_ID')
        start = self.attrs.get('START')
        end = self.attrs.get('END')
        donor_start = self.attrs.get('DONOR_START')
        donor_end = self.attrs.get('DONOR_END')
        left_insert_start = self.attrs.get('LEFT_INSERT_START')
        left_insert_end = self.attrs.get('LEFT_INSERT_END')
        right_insert_start = self.attrs.get('RIGHT_INSERT_START')
        right_insert_end = self.attrs.get('RIGHT_INSERT_END')
        return hash((self.location.start, self.location.end, self.ref, self.alt,
            self.type, donor_tx_id, start, end, donor_start, donor_end,
            left_insert_start, left_insert_end, right_insert_start, right_insert_end))

    def __repr__(self) -> str:
        """Return representation of the VEP record."""
        return f"{self.location.start}:{self.location.end} {self.ref} ->" +\
            f" {self.alt}"

    def __eq__(self, other:VariantRecord) -> bool:
        """ equal to """
        return self.location == other.location and \
            self.ref == other.ref and \
            self.alt == other.alt and \
            self.type == other.type

    def __ne__(self, other:VariantRecord) -> bool:
        """ not equal to """
        return not self == other

    def __gt__(self, other:VariantRecord) -> bool:
        """ greater than """
        if self.location > other.location:
            return True
        if self.location == other.location:
            if self.alt > other.alt:
                return True
            if self.ref == other.ref:
                return self.type > other.type
        return False

    def __ge__(self, other:VariantRecord) -> bool:
        """ greather or equal to """
        return self == other or self > other

    def __lt__(self, other:VariantRecord) -> bool:
        """ less then """
        return not self >= other

    def __le__(self, other:VariantRecord) -> bool:
        """ less or equal to """
        return not self > other

    @property
    def transcript_id(self) -> str:
        """ Transcript ID """
        if 'TRANSCRIPT_ID' in self.attrs:
            return self.attrs['TRANSCRIPT_ID']
        return self.location.seqname

    @property
    def accepter_transcript_id(self) -> str:
        """ accetper transcript ID """
        return self.attrs['ACCEPTER_TRANSCRIPT_ID']

    def get_donor_start(self) -> int:
        """ Get donor start position """
        if self.type in ['Insertion', 'Substitution']:
            return int(self.attrs['DONOR_START'])
        raise ValueError(f"Don't know how to get donor start for variant type "
            f"{self.type}")

    def get_donor_end(self) -> int:
        """ Get donor end position """
        if self.type in ['Insertion', 'Substitution']:
            return int(self.attrs['DONOR_END'])
        raise ValueError(f"Don't know how to get donor start for variant type "
            f"{self.type}")

    def get_accepter_position(self) -> int:
        """ Get accepter position for fusion only """
        if self.type != 'Fusion':
            raise ValueError("Don't know how to get accetoer start for "
                f"variant type {self.type}")
        return int(self.attrs['ACCEPTER_POSITION'])

    def get_ref_len(self) -> int:
        """ Get the length of the ref """
        if not self.alt.startswith('<'):
            return len(self.ref)
        if self.type == 'Fusion':
            return 1
        if self.type == 'Insertion':
            return 0
        if self.type in ['Deletion', 'Substitution']:
            return self.location.end - self.location.start
        raise ValueError(f"Don't know how to get ref len for variant type "
            f"{self.type}")

    def get_alt_len(self) -> int:
        """ Get the length of the alt """
        if not self.alt.startswith('<'):
            return len(self.alt)
        if self.type == 'Fusion':
            return 1
        if self.type == 'Deletion':
            return 1
        if self.type in ['Insertion', 'Substitution']:
            return self.get_donor_end() - self.get_donor_start()
        raise ValueError(f"Don't know how to get alt len for variant type "
            f"{self.type}")

    def to_string(self) -> str:
        """ Convert to a GVF record. """
        chrom = self.location.seqname
        # using 1-base position
        pos = str(int(self.location.start) + 1)
        _id = self.id
        qual = '.'
        _filter = '.'

        if self.type in constant.SINGLE_NUCLEOTIDE_SUBSTITUTION:
            ref = str(self.ref)
            alt = str(self.alt)
        elif self.type == 'Fusion':
            ref = str(self.ref[0])
            alt = '<FUSION>'
        elif self.type in ['Insertion', 'Deletion', 'Substitution']:
            ref = str(self.ref[0])
            alt = f'<{self.type.upper()[:3]}>'
        else:
            ref = str(self.ref[0])
            alt = f'<{self.type.upper()}>'

        info = self.info
        return '\t'.join([chrom, pos, _id, ref, alt, qual, _filter, info])

    @property
    def info(self) -> str:
        """ Get property of the INFO field """
        out = ''
        for key,val in self.attrs.items():
            # using 1-base position
            if key in constant.ATTRS_POSITION:
                val = str(int(val) + 1)
            elif isinstance(val, list):
                val = ','.join([str(x) for x in val])
            out += f'{key.upper()}={val};'
        return out.rstrip(';')

    def is_snv(self) -> bool:
        """ Checks if the variant is a single nucleotide variant. """
        return len(self.ref) == 1 and len(self.alt) == 1

    def is_indel(self) -> bool:
        """ Checks if the variant is an indel """
        return not self.alt.startswith('<') and len(self.ref) != len(self.alt)

    def is_insertion(self) -> bool:
        """ Checks if the variant is an insertion """
        if self.type == 'Insertion':
            return True
        return len(self.ref) == 1 and not self.alt.startswith('<') and len(self.alt) > 1

    def is_deletion(self) -> bool:
        """ Checks if the variant is a deletion. """
        if self.type == 'Deletion':
            return True
        return len(self.ref) > 1 and len(self.alt) == 1

    def is_fusion(self) -> bool:
        """ Check if this is a fusion """
        return self.type == 'Fusion'

    def is_alternative_splicing(self) -> bool:
        """ Check if this is an alternative splicing event """
        return any(self.id.startswith(x) for x in constant.RMATS_TYPES)

    def is_codon_reassignment(self) -> bool:
        """ Check if the variant is a codon reassignment """
        return self.type in constant.CODON_REASSIGNMENTS_TYPES

    def is_merged_mnv(self) -> bool:
        """ Check if the variant is a MNV merged from individual adjacent variants. """
        return self.attrs.get('MERGED_MNV', False)

    def is_in_frame_fusion(self, anno:GenomicAnnotation):
        """ Check if this is a in-frame fusion. A in-frame fusion is only when
        both donor and accepter transcripts a protein coding (which known
        reading frame), and the sum of donor and accepter breakpoint reading
        frame offset equals to 0 or 3. """
        if not self.is_fusion():
            raise ValueError(f'Variant {self.id} is not fusion.')
        donor_model = anno.transcripts[self.transcript_id]
        accepter_model = anno.transcripts[self.attrs['ACCEPTER_TRANSCRIPT_ID']]
        if not donor_model.is_protein_coding or not accepter_model.is_protein_coding:
            return False
        if 'TRANSCRIPT_ID' in self.attrs:
            raise ValueError('Variant must be in transcriptional coordinate.')
        donor_breakpoint = self.location.start
        donor_offset = (donor_breakpoint - donor_model.get_cds_start_index()) % 3

        accepter_breakpoint = anno.coordinate_gene_to_transcript(
            index=self.attrs['ACCEPTER_TRANSCRIPT_ID'],
            gene=anno.transcripts[self.attrs['ACCEPTER_GENE_ID']],
            transcript=anno.transcripts[self.attrs['ACCEPTER_TRANSCRIPT_ID']]
        )
        accepter_cds_start = accepter_model.get_cds_start_index()
        accepter_offset = (accepter_breakpoint - accepter_cds_start) % 3
        return (donor_offset + accepter_offset) % 3 == 0

    def is_circ_rna(self) -> bool:
        """ check if this is a circRNA """
        return self.type == 'circRNA'

    def is_frameshifting(self) -> bool:
        """ Checks if the variant is frameshifting. """
        if self.type == 'Fusion':
            return True
        ref_len = len(self.location)
        if self.type in ['Insertion', 'Substitution']:
            end = self.get_donor_end()
            start = self.get_donor_start()
            alt_len = end - start
            if self.type == 'Insertion':
                alt_len += 1
        elif self.type == 'Deletion':
            alt_len = 1
        else:
            alt_len = len(self.alt)
        return (ref_len - alt_len) % 3 != 0

    def is_inframe_indel(self) -> bool:
        """ checks if it is a inframe indel """
        return self.type in {'INDEL', 'Insertion', 'Deletion', 'Substitution'} \
            and not self.is_frameshifting()

    def set_end_inclusion(self):
        """ Set end inclusion to True """
        self.attrs['END_INCLUSION'] = True

    def unset_end_inclusion(self):
        """ Set end inclusion to False """
        self.attrs['END_INCLUSION'] = False

    def is_end_inclusion(self) -> bool:
        """ Is end inclusion """
        return self.attrs.get('END_INCLUSION') is True

    def frames_shifted(self):
        """ get number of nucleotide shifted
        TODO: tests needed """
        if self.type == 'Fusion':
            return 0
        ref_len = len(self.location)
        if self.type in ['Insertion', 'Substitution']:
            end = self.get_donor_end()
            start = self.get_donor_start()
            alt_len = end - start
            if self.type == 'Insertion':
                alt_len += 1
        elif self.type == 'Deletion':
            alt_len = 1
        else:
            alt_len = len(self.alt)
        return (ref_len - alt_len) % 3

    def is_spanning_over_splicing_site(self, anno:GenomicAnnotation,
            transcript_id:str) -> bool:
        """ Check if this is spanning over splicing site """
        gene_id = self.location.seqname
        if gene_id not in anno.genes:
            raise ValueError(ERROR_VARIANT_NOT_IN_GENE_COORDINATE)
        if self.is_fusion():
            start = self.location.start - 1
            end = self.location.end - 1
        else:
            start = self.location.start
            end = self.location.end

        try:
            anno.coordinate_gene_to_transcript(start, gene_id, transcript_id)
            start_in_intron = False
        except ValueError as e:
            if e.args[0] == ERROR_INDEX_IN_INTRON:
                start_in_intron = True
            else:
                raise e
        try:
            anno.coordinate_gene_to_transcript(end - 1, gene_id, transcript_id)
            end_in_intron = False
        except ValueError as e:
            if e.args[0] == ERROR_INDEX_IN_INTRON:
                end_in_intron = True
            else:
                raise e
        return ((not start_in_intron) and end_in_intron) \
                or (start_in_intron and (not end_in_intron))

    def to_transcript_variant(self, anno:GenomicAnnotation, genome:DNASeqDict,
            tx_id:str=None, cached_seqs:Dict[str, DNASeqRecordWithCoordinates]=None
            ) -> VariantRecord:
        """ Get variant records with transcription coordinates """
        if cached_seqs is None:
            cached_seqs = {}
        if not tx_id:
            tx_id = self.attrs['TRANSCRIPT_ID']
        if not tx_id:
            raise ValueError(ERROR_NO_TX_AVAILABLE)
        tx_model = anno.transcripts[tx_id]
        chrom = tx_model.transcript.chrom
        if tx_id not in cached_seqs:
            tx_seq = tx_model.get_transcript_sequence(genome[chrom])
            cached_seqs[tx_id] = tx_seq
        else:
            tx_seq = cached_seqs[tx_id]

        gene_id = self.location.seqname
        strand = tx_model.transcript.strand

        tx_start = tx_model.transcript.location.start if strand == 1 else \
            tx_model.transcript.location.end - 1

        var_start = self.location.start
        var_end = self.location.end

        start_genomic = anno.coordinate_gene_to_genomic(var_start, gene_id)
        end_genomic = anno.coordinate_gene_to_genomic(var_end - 1, gene_id)
        end_genomic += 1

        variant_start_before_tx_start = \
            (strand == 1 and start_genomic < tx_start) or \
            (strand == -1 and tx_start < start_genomic)
        variant_end_before_tx_start = \
            (strand == 1 and end_genomic <= tx_start) or \
            (strand == -1 and tx_start <= end_genomic)

        if variant_start_before_tx_start:
            if variant_end_before_tx_start:
                raise ValueError(
                    'Variant not associated with the given transcript'
                )
            start = 0
            end = anno.coordinate_gene_to_transcript(var_end, gene_id, tx_id) + 1
            ref = str(tx_seq.seq[0:end])
            alt = str(self.alt[1:] + ref[-1])
        else:
            if self.is_fusion():
                start = anno.coordinate_gene_to_transcript(var_start - 1, gene_id, tx_id) + 1
                end = start + 1
                # if start == len(tx_seq):
                #     raise err.FusionBreakpointIsEndOfTranscript(self.id)
            else:
                start = anno.coordinate_gene_to_transcript(var_start, gene_id, tx_id)
                end = anno.coordinate_gene_to_transcript(var_end - 1, gene_id, tx_id) + 1
            ref = self.ref
            alt = self.alt
            # If the distance from start to end on the transcript does not
            # equal to that on the gene, the variant contains at least one
            # intron. The ref is then regenerated.
            if end - start != len(self.location):
                ref = str(tx_seq.seq[start:end])
        location = FeatureLocation(seqname=tx_id, start=start, end=end)
        attrs = copy.deepcopy(self.attrs)
        del attrs['TRANSCRIPT_ID']
        attrs['GENE_ID'] = self.location.seqname
        return VariantRecord(
            location=location,
            ref=ref,
            alt=alt,
            _type=self.type,
            _id=self.id,
            attrs=attrs
        )

    def has_transcript(self):
        """ Checks if the variant has any transcript IDs annotated """
        return 'TRANSCRIPT_ID' in self.attrs

    def is_stop_lost(self, stop:int):
        """ Check if this is a stop lost mutation """
        loc = FeatureLocation(start=stop, end=stop+3)
        return self.location.overlaps(loc)

    def to_end_inclusion(self, seq:DNASeqRecord):
        """ Convert the variant to start exlusion and end inclusion format """
        if self.type != 'Insertion' and self.alt.startswith('<'):
            raise ValueError(
                f'This variant should not be converted to end inclusion: {self}'
            )
        location = FeatureLocation(
            seqname=self.location.seqname,
            start=self.location.start + 1,
            end=self.location.end + 1
        )
        self.set_end_inclusion()
        if self.type == 'Insertion':
            ref = self.ref[1:] + str(seq.seq[self.location.end])
            self.location = location
            self.ref = ref
        else:
            ref = self.ref[1:] + str(seq.seq[self.location.end])
            alt = self.alt[1:] + str(seq.seq[self.location.end])
            self.location = location
            self.ref = ref
            self.alt = alt

    def shift_breakpoint_to_closest_exon(self, anno:GenomicAnnotation):
        """ Shift fusion breakpoints to the closest exon. Donor breakpoint will
        shift to the upstream and accepter breakpoint to the downstream. """
        if not self.is_fusion():
            raise ValueError(
                "Don't know how to shift breakpoint for non-fusion variants."
            )
        donor_gene_id = self.location.seqname
        donor_tx_id = self.attrs['TRANSCRIPT_ID']
        donor_tx_model = anno.transcripts[donor_tx_id]
        left_breakpoint = anno.coordinate_gene_to_genomic(
            index=self.location.start - 1, gene=donor_gene_id
        )
        if donor_tx_model.is_exonic(left_breakpoint):
            left_insertion_start = None
            left_insertion_end = None
        else:
            upstream_exon_end = donor_tx_model.get_upstream_exon_end(left_breakpoint)
            left_insertion_start = anno.coordinate_genomic_to_gene(
                index=upstream_exon_end, gene=donor_gene_id
            ) + 1
            left_insertion_end = self.location.start
            self.location = FeatureLocation(
                start=left_insertion_start,
                end=left_insertion_start + 1,
                seqname = donor_gene_id
            )

        accepter_gene_id = self.attrs['ACCEPTER_GENE_ID']
        accepter_tx_id = self.attrs['ACCEPTER_TRANSCRIPT_ID']
        accepter_tx_model = anno.transcripts[accepter_tx_id]
        right_breakpoint = anno.coordinate_gene_to_genomic(
            index=self.get_accepter_position(), gene=accepter_gene_id
        )
        if accepter_tx_model.is_exonic(right_breakpoint):
            right_insertion_start = None
            right_insertion_end = None
        else:
            downstream_exon_start = accepter_tx_model.get_downstream_exon_start(
                pos=right_breakpoint
            )
            right_insertion_start = self.get_accepter_position()
            right_insertion_end = anno.coordinate_genomic_to_gene(
                downstream_exon_start, accepter_gene_id
            )
            self.attrs['ACCEPTER_POSITION'] = anno.coordinate_genomic_to_gene(
                index=downstream_exon_start, gene=accepter_gene_id
            )

        self.attrs.update({
            'LEFT_INSERTION_START': left_insertion_start,
            'LEFT_INSERTION_END': left_insertion_end,
            'RIGHT_INSERTION_START': right_insertion_start,
            'RIGHT_INSERTION_END': right_insertion_end
        })

    def shift_deletion_up(self, tx_seq:DNASeqRecord):
        """ shift deletion variant up exact one position """
        location = FeatureLocation(
            start=self.location.start - 1,
            end=self.location.end,
            seqname=self.location.seqname
        )
        ref = str(tx_seq.seq[location.start])
        self.location = location
        self.ref = ref



r""" `parseCIRCExplorer` takes the identified circRNA results from
[CIRCexplorer](https://circexplorer2.readthedocs.io/) and save as a
GVF file. The GVF file can be later used to call variant peptides using
[callVariant](call-variant.md). Noted that only known circRNA is supported (
\*_circular_known.txt) """
from __future__ import annotations
import argparse
from typing import List, Dict
from pathlib import Path
from moPepGen import get_logger, circ, err
from moPepGen.parser import CIRCexplorerParser
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.tsv', '.txt']
OUTPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def add_subparser_parse_circexplorer(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen parseCIRCexplorer """
    p = subparsers.add_parser(
        name='parseCIRCexplorer',
        help='Parse CIRCexplorer result',
        description='Parse CIRCexplorer result to a TSV format for moPepGen to'
        ' call variant peptides',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS,
        message="File path to CIRCexplorer's TSV output."
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    p.add_argument(
        '--circexplorer3',
        action='store_true',
        help='Using circRNA resutls from CIRCexplorer3'
    )
    p.add_argument(
        '--min-read-number',
        type=int,
        help='Minimal number of junction read counts.',
        default=1,
        metavar='<number>'
    )
    p.add_argument(
        '--min-fpb-circ',
        type=float,
        help='Minimal CRICscore value for CIRCexplorer3. Recommends to 1,'
        ' defaults to None',
        default=None,
        metavar='<number>'
    )
    p.add_argument(
        '--min-circ-score',
        type=float,
        help='Minimal CIRCscore value for CIRCexplorer3. Recommends to 1,'
        ' defaults to None',
        default=None,
        metavar='<number>'
    )
    p.add_argument(
        '--intron-start-range',
        type=str,
        help='The range of difference allowed between the intron start and'
        ' the reference position.',
        default='-2,0',
        metavar='<number>'
    )
    p.add_argument(
        '--intron-end-range',
        type=str,
        help='The range of difference allowed between the intron end and'
        ' the reference position.',
        default='-100,5',
        metavar='<number>'
    )
    common.add_args_source(p)
    common.add_args_reference(p, genome=False, proteome=False)
    common.add_args_debug_level(p)
    p.set_defaults(func=parse_circexplorer)
    common.print_help_if_missing_args(p)
    return p

def parse_circexplorer(args:argparse.Namespace):
    """ Parse circexplorer known circRNA results. """
    logger = get_logger()

    input_path:Path = args.input_path
    output_path:Path = args.output_path
    common.validate_file_format(
        input_path, INPUT_FILE_FORMATS, check_readable=True
    )
    common.validate_file_format(
        output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    intron_start_range = common.parse_range(args.intron_start_range)
    intron_end_range = common.parse_range(args.intron_end_range)

    common.print_start_message(args)

    _, anno, *_ = common.load_references(args, False, False)

    circ_records:Dict[str, List[circ.CircRNAModel]] = {}

    for record in CIRCexplorerParser.parse(input_path, args.circexplorer3):
        if not args.circexplorer3:
            if not record.is_valid(args.min_read_number):
                continue
        elif not record.is_valid(args.min_read_number, args.min_fbr_circ, \
                args.min_circ_score):
            continue
        try:
            circ_record = record.convert_to_circ_rna(anno, intron_start_range,
                intron_end_range)
        except err.ExonNotFoundError:
            logger.warning(
                "The CIRCexplorer record % from transcript %s contains an unknown exon."
                " Skipping it from parsing.",
                record.name, record.isoform_name
            )
            continue
        except err.IntronNotFoundError:
            logger.warning(
                "The CIRCexplorer record %s from transcript %s contains an unknown"
                " intron. Skipping it from parsing.",
                record.name, record.isoform_name
            )
            continue
        except:
            logger.error('Exception raised from record: %s', record.name)
            raise
        gene_id = circ_record.gene_id
        if gene_id not in circ_records:
            circ_records[gene_id] = []
        circ_records[gene_id].append(circ_record)

    if not circ_records:
        logger.warning('No variant record is saved.')
        return

    genes_rank = anno.get_genes_rank()
    ordered_keys = sorted(circ_records.keys(), key=lambda x:genes_rank[x])

    records = []
    for key in ordered_keys:
        val = circ_records[key]
        records.extend(val)

    metadata = common.generate_metadata(args)

    with open(output_path, 'w') as handle:
        circ.io.write(records, metadata, handle)

    logger.info("CircRNA records written to disk.")



""" `parseFusionCatcher` takes the identified fusion transcript results from
[FusionCatcher](https://github.com/ndaniel/fusioncatcher) and save as a
GVF file. The GVF file can be later used to call variant peptides using
[callVariant](call-variant.md)."""
from typing import List
from pathlib import Path
import argparse
from moPepGen import get_logger, seqvar, parser, err
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.tsv', '.txt']
OUTPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def add_subparser_parse_fusion_catcher(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen parseFusionCatcher """

    p = subparsers.add_parser(
        name='parseFusionCatcher',
        help='Parse FusionCatcher result for moPepGen to call variant peptides.',
        description='Parse the FusionCatcher result to GVF format of variant'
        ' records for moPepGen to call variant peptides. The genome',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS,
        message="File path to FusionCatcher's output TSV file."
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    p.add_argument(
        '--max-common-mapping',
        type=int,
        help='Maximal number of common mapping reads.',
        metavar='<number>',
        default=0
    )
    p.add_argument(
        '--min-spanning-unique',
        help='Minimal spanning unique reads.',
        type=int,
        default=5,
        metavar='<number>'
    )
    common.add_args_source(p)
    common.add_args_reference(p, proteome=False)
    common.add_args_debug_level(p)
    p.set_defaults(func=parse_fusion_catcher)
    common.print_help_if_missing_args(p)
    return p

def parse_fusion_catcher(args:argparse.Namespace) -> None:
    """ Parse FusionCatcher output and save it in GVF format. """
    logger = get_logger()
    # unpack args
    fusion = args.input_path
    output_path:Path = args.output_path
    common.validate_file_format(
        fusion, INPUT_FILE_FORMATS, check_readable=True
    )
    common.validate_file_format(
        output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    common.print_start_message(args)

    genome, anno, *_ = common.load_references(args=args, load_canonical_peptides=False)

    variants:List[seqvar.VariantRecord] = []

    for record in parser.FusionCatcherParser.parse(fusion):
        if record.counts_of_common_mapping_reads > args.max_common_mapping:
            continue
        if record.spanning_unique_reads < args.min_spanning_unique:
            continue
        try:
            var_records = record.convert_to_variant_records(anno, genome)
        except err.GeneNotFoundError:
            continue
        variants.extend(var_records)

    logger.info('FusionCatcher output %s loaded.', fusion)

    if not variants:
        logger.warning('No variant record is saved.')
        return

    genes_rank = anno.get_genes_rank()
    variants = sorted(variants, key=lambda x: genes_rank[x.location.seqname])

    logger.info('Variants sorted.')

    metadata = common.generate_metadata(args)

    seqvar.io.write(variants, output_path, metadata)

    logger.info("Variants written to disk.")



""" Module for GenomicAnnotationOnDisk """
from __future__ import annotations
from typing import Dict, IO, Union, Tuple
import io
from pathlib import Path
from moPepGen.version import MetaVersion
from moPepGen.gtf.GenomicAnnotation import GenomicAnnotation
from moPepGen.gtf.GTFPointer import (
    GenePointerDict, TranscriptPointerDict, iterate_pointer,
    GenePointer, TranscriptPointer
)


class GenomicAnnotationOnDisk(GenomicAnnotation):
    """ An on disk version of GenomicAnnotation that uses minimal memory. """
    def __init__(self, genes:GenePointerDict=None,
            transcripts:TranscriptPointerDict=None,
            source:str=None):
        """ Constructor """
        # pylint: disable=W0231
        self.genes = genes or GenePointerDict()
        self.transcripts = transcripts or TranscriptPointerDict()
        self.source = source
        self.gene_id_version_mapper:Dict[str,str] = None
        self.version = MetaVersion()
        self._cached_tx_seqs = []
        self.handle = None

    def __del__(self):
        """ Hook for deconstruction """
        if self.handle:
            self.handle.close()

    def dump_gtf(self, *args, **kwargs) -> None:
        """ Forbidden """
        raise NotImplementedError

    def infer_source(self):
        """ Infer GTF source (GENCODE vs ENSEMBL) from data. """
        i = 0
        inferred = {}
        for tx in self.transcripts.keys():
            i += 1
            if i > 100:
                break
            s = self.transcripts.get_pointer(tx).source
            if s in inferred:
                inferred[s] += 1
            else:
                inferred[s] = 1
        self.source = sorted(inferred.items(), key=lambda it: it[1])[-1][0]

    def init_handle(self, handle:Union[str, IO, Path]):
        """ Initiate file handle """
        # pylint: disable=R1732
        if isinstance(handle, str):
            handle = Path(handle)

        if isinstance(handle, Path):
            if handle.suffix.lower() != '.gtf':
                raise ValueError(f"Unsupported file formate of {handle}")
            ihandle = open(handle, 'rb')
        elif isinstance(handle, io.IOBase):
            ihandle = handle
        else:
            raise ValueError(f"handle {handle} of type {handle.__class__} is not supported.")

        self.handle = ihandle

    @staticmethod
    def get_index_files(file:Union[str,Path]) -> Tuple[Path,Path]:
        """ Get index file paths given the GTF file path. """
        if isinstance(file, str):
            file = Path(file)
        basename = str(file.name).split('.gtf', maxsplit=1)[0]
        gene_idx_file = file.parent/f"{basename}_gene.idx"
        tx_idx_file = file.parent/f"{basename}_tx.idx"

        return gene_idx_file, tx_idx_file

    def generate_index(self, handle:Union[IO, str, Path], source:str=None):
        """ Generate GTF index """
        self.init_handle(handle)

        for pointer in iterate_pointer(self.handle, source):
            if isinstance(pointer, GenePointer):
                self.genes[pointer.key] = pointer
            else:
                self.transcripts[pointer.key] = pointer

        if source:
            self.source = source
        else:
            self.infer_source()

    def load_index(self, file:Union[str,Path], source:str):
        """ Load index from idx files """
        self.init_handle(file)
        gene_idx_file, tx_idx_file = self.get_index_files(file)

        if not gene_idx_file.exists():
            raise ValueError(f"Gene index file cannot be found in {file.parent}")
        if not tx_idx_file.exists():
            raise ValueError(f"Transcript index file cannot be found in {file.parent}")

        with open(gene_idx_file, 'rt') as handle:
            for line in handle:
                if line.startswith('#'):
                    continue
                fields = line.rstrip().split('\t')
                pointer = GenePointer(
                    self.handle, key=fields[0],
                    start=int(fields[1]), end=int(fields[2]),
                    source=source, transcripts=fields[3].split(',')
                )
                self.genes[pointer.key] = pointer

        with open(tx_idx_file, 'rt') as handle:
            for line in handle:
                if line.startswith('#'):
                    continue
                fields = line.rstrip().split('\t')
                is_protein_coding = {'None': None, 'True': True, 'False': False}[fields[3]]
                pointer = TranscriptPointer(
                    handle=self.handle, key=fields[0],
                    start=int(fields[1]), end=int(fields[2]),
                    source=source, is_protein_coding=is_protein_coding
                )
                self.transcripts[pointer.key] = pointer



""" Validate callVariant result with bruteForce """
import argparse
from contextlib import redirect_stdout
from pathlib import Path
from typing import List
import sys
from Bio import SeqIO
from moPepGen.cli import call_variant_peptide, common
from moPepGen import get_logger
from moPepGen.util import downsample_reference, brute_force


INPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ parse args """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='validateVariantCalling',
        help='Validate the variant peptide calling result of the graph-based'
        ' algorithm with the brute force algorithm.'
    )
    common.add_args_input_path(
        parser, INPUT_FILE_FORMATS,
        plural=True,
        message='File path to the GVF files generated by moPepGen parsers.'
    )
    parser.add_argument(
        '--tx-id',
        type=str,
        help='Transcript ID',
        metavar='<value>',
        nargs='+'
    )
    parser.add_argument(
        '-o', '--output-dir',
        type=Path,
        help='Path to the output diractory.',
        metavar='<file>'
    )
    parser.add_argument(
        '-f', '--force',
        action='store_true',
        help='If not set, the program stops when there are more than 10'
        ' variants. When this flag is set, the program runs anyway. Noted that '
        ' the runtime is going to increase quickly after 10 variants.',
        default=False
    )
    parser.add_argument(
        '--variant-ids',
        type=str,
        help='List of variant labels.',
        nargs='*',
        metavar='<values>'
    )
    common.add_args_reference(parser, proteome=True, index=False)
    parser.set_defaults(func=main)
    common.print_help_if_missing_args(parser)
    return parser

def call_downsample_reference(genome:Path, anno:Path, protein:Path, tx_id:List[str],
        output_dir:Path):
    """ downsample reference """
    args = argparse.Namespace()
    args.genome_fasta = genome
    args.annotation_gtf = anno
    args.proteome_fasta = protein
    args.tx_list = tx_id
    args.gene_list = None
    args.output_dir = output_dir
    args.miscleavage = 2
    args.min_mw = 500.
    args.translate_noncoding = 'false'
    downsample_reference.main(args)

def extract_gvf(tx_id:List[str], gvf_files:List[Path], output_dir:Path) -> List[Path]:
    """ extract GVF """
    i = 0
    temp_gvfs = []
    for j, gvf_file in enumerate(gvf_files):
        output_path = output_dir/f"{'-'.join(tx_id)}_{j}_{gvf_file.name}"
        temp_gvfs.append(output_path)
        with open(gvf_file, 'rt') as in_handle, open(output_path, 'wt') as out_handle:
            for line in in_handle:
                if line.startswith('#'):
                    out_handle.write(line)
                elif any(x in line for x in tx_id):
                    out_handle.write(line)
                    i += 0
    if i > 10:
        get_logger().info(
            "%i variants found. The brute force caller is going to take a while.",
            i
        )
    return temp_gvfs

def call_variant(gvf_files:Path, ref_dir:Path, output_fasta:Path, graph_output_dir:Path):
    """ call variant """
    args = argparse.Namespace()
    args.index_dir = None
    args.command = 'callPeptides'
    args.input_path = gvf_files
    args.genome_fasta = ref_dir/'genome.fasta'
    args.annotation_gtf = ref_dir/'annotation.gtf'
    args.proteome_fasta = ref_dir/'proteome.fasta'
    args.reference_source = None
    args.max_adjacent_as_mnv = 2
    args.backsplicing_only = False
    args.coding_novel_orf = False
    args.selenocysteine_termination = True
    args.w2f_reassignment = True
    args.invalid_protein_as_noncoding = False
    args.output_path = output_fasta
    args.graph_output_dir = graph_output_dir
    args.quiet = False
    args.debug_level = 1
    args.cleavage_rule = 'trypsin'
    args.cleavage_exception = None
    args.miscleavage = 2
    args.min_mw = 500.
    args.min_length = 7
    args.max_length = 25
    args.max_variants_per_node = 9
    args.additional_variants_per_misc = 2
    args.min_nodes_to_collapse = 30
    args.naa_to_collapse = 5
    args.noncanonical_transcripts = False
    args.threads = 1
    call_variant_peptide(args=args)

def call_brute_force(gvf_files:Path, ref_dir:Path, output_path:str, force:bool,
        variant_ids:List[str]):
    """ call brute force """
    args = argparse.Namespace()
    args.input_gvf = gvf_files
    args.reference_dir = ref_dir
    args.force = force
    args.variant_ids = variant_ids
    args.max_adjacent_as_mnv = 2
    args.selenocysteine_termination = True
    args.w2f_reassignment = True
    args.cleavage_rule = 'trypsin'
    args.cleavage_exception = None
    args.miscleavage = 2
    args.min_mw = 500.
    args.min_length = 7
    args.max_length = 25

    with open(output_path, 'wt') as handle:
        with redirect_stdout(handle):
            brute_force.main(args)

def assert_equal(variant_fasta:Path, brute_force_txt:Path, output_dir:Path):
    """ assert equal """
    logger = get_logger()
    with open(variant_fasta, 'rt') as handle:
        variant_seqs = set()
        for seq in SeqIO.parse(handle, 'fasta'):
            variant_seqs.add(str(seq.seq))
    with open(brute_force_txt, 'rt') as handle:
        brute_force_seqs = set()
        for line in handle:
            brute_force_seqs.add(line.rstrip())

    if variant_seqs != brute_force_seqs:
        logger.info('Not equal!')
        variant_only = variant_seqs - brute_force_seqs
        brute_force_only = brute_force_seqs - variant_seqs
        if variant_only:
            variant_only_file = output_dir/'call_variant_only.txt'
            with open(variant_only_file, 'wt') as handle:
                for seq in variant_only:
                    handle.write(seq + '\n')
        if brute_force_only:
            brute_force_only_file = output_dir/'brute_force_only.txt'
            with open(brute_force_only_file, 'wt') as handle:
                for seq in brute_force_only:
                    handle.write(seq + '\n')
        sys.exit(1)
    logger.info('Equal!')

def main(args:argparse.Namespace):
    """ main entrypoint """
    logger = get_logger()
    for file in args.input_path:
        common.validate_file_format(file, INPUT_FILE_FORMATS, True)
    output_dir:Path = args.output_dir
    output_dir.mkdir(exist_ok=True)
    ref_dir = output_dir/'index'
    ref_dir.mkdir(exist_ok=True)
    call_downsample_reference(
        genome=args.genome_fasta,
        anno=args.annotation_gtf,
        protein=args.proteome_fasta,
        tx_id=args.tx_id,
        output_dir=ref_dir
    )

    logger.info('Reference files downsampling completed.')

    temp_gvfs = extract_gvf(
        tx_id=args.tx_id,
        gvf_files=args.input_path,
        output_dir=output_dir
    )

    logger.info('Transcript variants extracted from GVF.')

    variant_fasta = output_dir/'call_variant.fasta'
    graph_dir = output_dir/'graph'
    graph_dir.mkdir(exist_ok=True)
    call_variant(
        gvf_files=temp_gvfs,
        ref_dir=ref_dir,
        output_fasta=variant_fasta,
        graph_output_dir=graph_dir
    )

    logger.info('Variant peptide calling completed.')

    brute_force_txt = output_dir/'brute_force.txt'
    call_brute_force(
        gvf_files=temp_gvfs,
        ref_dir=ref_dir,
        output_path=brute_force_txt,
        force=args.force,
        variant_ids=args.variant_ids
    )

    logger.info('Brute force completed.')

    assert_equal(variant_fasta, brute_force_txt, output_dir)



""" Module for STAR-Fusion parser """
from __future__ import annotations
from typing import Iterable, List
import itertools
from moPepGen.SeqFeature import FeatureLocation
from moPepGen import seqvar, gtf, dna, err


def parse(path:str) -> Iterable[STARFusionRecord]:
    """ Parse the STAR-Fusion's output and returns an iterator.

    Args:
        path (str): Path to the STAR-Fusion's output file.
    """
    with open(path, 'r') as handle:
        line = next(handle, None)
        while line:
            if line.startswith('#'):
                line = next(handle, None)
                continue
            fields = line.rstrip().split('\t')
            yield STARFusionRecord(
                fusion_name=fields[0],
                junction_read_count=int(fields[1]),
                spanning_frag_count=int(fields[2]),
                est_j=float(fields[3]),
                est_s=float(fields[4]),
                splice_type=fields[5],
                left_gene=fields[6].split('^')[-1],
                left_breakpoint=fields[7],
                right_gene=fields[8].split('^')[-1],
                right_breakpoint=fields[9],
                junction_reads=fields[10].split(','),
                spanning_frags=fields[11].split(','),
                large_anchor_support=fields[12],
                ffpm=float(fields[13]),
                left_break_dinuc=fields[14],
                left_break_entropy=float(fields[15]),
                right_break_dinuc=fields[16],
                right_break_entropy=float(fields[17]),
                annots=fields[18].strip('][').replace('"', '').split(',')
            )
            line = next(handle, None)


class STARFusionRecord():
    """ Defines a record of STAR-Fusion's output. """
    def __init__(self, fusion_name:str, junction_read_count:int,
            spanning_frag_count:int, est_j:float, est_s:float,
            splice_type:str, left_gene:str, left_breakpoint:str,
            right_gene:str, right_breakpoint:str, junction_reads:List[str],
            spanning_frags:List[str], large_anchor_support:str, ffpm:float,
            left_break_dinuc:str, left_break_entropy:float,
            right_break_dinuc:str, right_break_entropy:float, annots:List[str]
            ):
        """"""
        self.fusion_name = fusion_name
        self.junction_read_count = junction_read_count
        self.spanning_frag_count = spanning_frag_count
        self.est_j = est_j
        self.est_s = est_s
        self.splice_type = splice_type
        self.left_gene = left_gene
        self.left_breakpoint = left_breakpoint
        self.right_gene = right_gene
        self.right_breakpoint = right_breakpoint
        self.junction_reads = junction_reads
        self.spanning_frags = spanning_frags
        self.large_anchor_support = large_anchor_support
        self.ffpm = ffpm
        self.left_break_dinuc = left_break_dinuc
        self.left_break_entropy = left_break_entropy
        self.right_break_dinuc = right_break_dinuc
        self.right_break_entropy = right_break_entropy
        self.annots = annots

    @property
    def left_breakpoint_position(self) -> int:
        """ Get left breakpoint position """
        return int(self.left_breakpoint.split(':')[1])

    @property
    def right_breakpoint_position(self) -> int:
        """ Get right breakpoint position """
        return int(self.right_breakpoint.split(':')[1])

    def get_donor_transcripts(self, anno:gtf.GenomicAnnotation
            ) -> List[gtf.TranscriptAnnotationModel]:
        """ Get all possible donor transcripts """
        pos = self.left_breakpoint_position - 1
        return anno.get_transcripts_with_position(self.left_gene, pos)

    def get_accepter_transcripts(self, anno:gtf.GenomicAnnotation
            ) -> List[gtf.TranscriptAnnotationModel]:
        """ Get all possible accepter transcripts """
        pos = self.right_breakpoint_position - 1
        return anno.get_transcripts_with_position(self.right_gene, pos)

    def convert_to_variant_records(self, anno:gtf.GenomicAnnotation,
            genome:dna.DNASeqDict) -> List[seqvar.VariantRecord]:
        """ Convert a STAR-Fusion's record to VariantRecord.

        Args:
            anno (Dict[str, Dict[str, gtf.TranscriptAnnotationModel]]): The
                genome annotation. The first level key is the gene ID, and the
                second level key is the transcript ID.
            genome (dna.DNASeqDict): The genome sequence records.
        Returns:
            List of VariantRecord
        """
        try:
            donor_model = anno.genes[self.left_gene]
        except KeyError as error:
            raise err.GeneNotFoundError(self.left_gene) from error
        donor_gene_symbol = donor_model.gene_name
        donor_chrom = self.left_breakpoint.split(':')[0]
        left_breakpoint = self.left_breakpoint_position
        donor_genome_position = f'{donor_chrom}:{left_breakpoint}:{left_breakpoint}'
        donor_position = anno.coordinate_genomic_to_gene(left_breakpoint - 1, self.left_gene) + 1
        donor_transcripts = self.get_donor_transcripts(anno)

        try:
            accepter_model = anno.genes[self.right_gene]
        except KeyError as error:
            raise err.GeneNotFoundError(self.right_gene) from error
        accepter_gene_symbol = accepter_model.gene_name
        accepter_chrom = self.right_breakpoint.split(':')[0]
        right_breakpoint = self.right_breakpoint_position
        accepter_genome_position = f'{accepter_chrom}:{right_breakpoint}:{right_breakpoint}'
        accepter_position = anno.coordinate_genomic_to_gene(right_breakpoint - 1, self.right_gene)
        accepter_transcripts = self.get_accepter_transcripts(anno)
        records = []

        if donor_model.strand == 1:
            ref_seq = genome[donor_chrom].seq[left_breakpoint + 1]
        else:
            ref_seq = genome[donor_chrom]\
                .seq[left_breakpoint:left_breakpoint + 1]\
                .reverse_complement()
            ref_seq = str(ref_seq)

        perms = itertools.product(donor_transcripts, accepter_transcripts)
        for donor_tx, accepter_tx in perms:
            donor_tx_id = donor_tx.transcript.transcript_id
            accepter_tx_id = accepter_tx.transcript.transcript_id

            fusion_id = f'FUSION-{donor_tx_id}:{donor_position}'\
                f'-{accepter_tx_id}:{accepter_position}'

            location = FeatureLocation(
                seqname=self.left_gene,
                start=donor_position,
                end=donor_position + 1
            )
            attrs = {
                'TRANSCRIPT_ID': donor_tx_id,
                'GENE_SYMBOL': donor_gene_symbol,
                'GENOMIC_POSITION': donor_genome_position,
                'ACCEPTER_GENE_ID': self.right_gene,
                'ACCEPTER_TRANSCRIPT_ID': accepter_tx_id,
                'ACCEPTER_SYMBOL': accepter_gene_symbol,
                'ACCEPTER_POSITION': accepter_position,
                'ACCEPTER_GENOMIC_POSITION': accepter_genome_position
            }
            record = seqvar.VariantRecord(
                location=location,
                ref=ref_seq,
                alt='<FUSION>',
                _type='Fusion',
                _id=fusion_id,
                attrs=attrs
            )
            records.append(record)
        return records



""" `updateIndex` creates new canoincal peptide pool to a moPepGen index dir
created by `generateIndex`. This command is useful when you want to use a
different enzyme or cleavage settings (*e.g.*, `miscleavages`, `min_length`) with
the same reference genome, annotation, and proteome. """
from __future__ import annotations
import argparse
import sys
from pathlib import Path
from moPepGen import params, get_logger
from moPepGen.index import IndexDir
from moPepGen.cli import common


# pylint: disable=W0212
def add_subparser_update_index(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen generateIndex """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='updateIndex',
        help='Update moPepGen index',
        description='Update moPepGen index',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    p.add_argument(
        '--index-dir',
        type=Path,
        help='Path to the directory of index files generated by moPepGen'
        ' generateIndex. If given, --genome-fasta, --proteome-fasta and'
        ' --anntotation-gtf will be ignored.',
        metavar='<file>',
        nargs='?',
        default=None
    )
    p.add_argument(
        '-f', '--force',
        action='store_true',
        help='Force write data to index dir.'
    )
    common.add_args_cleavage(p)
    common.add_args_debug_level(p)
    p.set_defaults(func=update_index)
    common.print_help_if_missing_args(p)
    return p

def update_index(args:argparse.Namespace):
    """ update index """
    logger = get_logger()

    common.print_start_message(args)

    index_dir = IndexDir(args.index_dir)
    index_dir.validate_metadata()

    rule:str = args.cleavage_rule
    miscleavage:int = int(args.miscleavage)
    min_mw:float = float(args.min_mw)
    min_length:int = int(args.min_length)
    max_length:int = int(args.max_length)
    exception = args.cleavage_exception

    cleavage_params = params.CleavageParams(
        enzyme=rule, exception=exception, miscleavage=miscleavage,
        min_mw=min_mw, min_length = min_length, max_length = max_length
    )
    pool_exists = index_dir.metadata.get_canonical_pool(cleavage_params) is not None
    if pool_exists and not args.force:
        logger.error(
            'Canonical peptide pool already exists with the given parameters.'
        )
        sys.exit(1)

    # load reference data
    anno = index_dir.load_annotation()
    logger.info('Genomic annotation loaded.')
    proteome = index_dir.load_proteome()
    logger.info('Proteome loaded.')

    # create canonical peptide pool
    canonical_peptides = proteome.create_unique_peptide_pool(
        anno=anno, rule=rule, exception=exception, miscleavage=miscleavage,
        min_mw=min_mw, min_length = min_length, max_length = max_length
    )
    logger.info('Canoincal peptide pool generated.')
    index_dir.save_canonical_peptides(canonical_peptides, cleavage_params, override=args.force)
    logger.info('Canoincal peptide pool saved.')

    if not pool_exists:
        index_dir.save_metadata()
        logger.info('metadata.json updated.')



""" Module for moPepGen parsers from different variant sources. """
from moPepGen.parser.REDItoolsParser import REDItoolsRecord
from moPepGen.parser import REDItoolsParser, VEPParser, STARFusionParser, \
	FusionCatcherParser, RMATSParser, CIRCexplorerParser, ArribaParser



""" `parseVEP` takes the output of Ensembl's [Variant Effector Predictor
](https://uswest.ensembl.org/info/docs/tools/vep/index.html) (VEP) and
convert it into the GVF file format that moPepGen internally uses. The result
VEP file can then be parsed to moPepGen's `callVariant` subcommand to call for
variant peptide sequences.
"""
from __future__ import annotations
import argparse
import gzip
from typing import Dict, List
from pathlib import Path
from moPepGen.parser import VEPParser
from moPepGen.err import MNVParsingError, TranscriptionStopSiteMutationError, \
    TranscriptionStartSiteMutationError, warning
from moPepGen import seqvar, get_logger
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.tsv', '.txt', '.tsv.gz', '.txt.gz']
OUTPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def add_subparser_parse_vep(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen parseVEP """

    p:argparse.ArgumentParser = subparsers.add_parser(
        name='parseVEP',
        help='Parse VEP output for moPepGen to call variant peptides.',
        description='Parse VEP output tsv to the GVF format of variant records'
        ' for moPepGen to call variant peptides. The genome assembly FASTA and'
        ' annotation GTF must come from the same GENCODE/ENSEMBL version, and'
        ' must the consistent with the VEP output.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS, plural=True,
        message='File path to the VEP output TXT file.'
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    common.add_args_source(p)
    common.add_args_reference(p, proteome=False)
    common.add_args_debug_level(p)
    p.set_defaults(func=parse_vep)
    common.print_help_if_missing_args(p)
    return p

def parse_vep(args:argparse.Namespace) -> None:
    """ Main entry point for the VEP parser. """
    logger = get_logger()
    # unpack args
    vep_files:List[Path] = args.input_path
    for file in vep_files:
        common.validate_file_format(
            file, INPUT_FILE_FORMATS, check_readable=True
        )
    output_path:Path = args.output_path
    common.validate_file_format(
        output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    common.print_start_message(args)

    genome, anno, *_ = common.load_references(args, load_canonical_peptides=False)

    vep_records:Dict[str, List[seqvar.VariantRecord]] = {}

    for vep_file in vep_files:
        opener = gzip.open if vep_file.suffix == '.gz' else open
        with opener(vep_file, 'rt') as handle:
            for record in VEPParser.parse(handle):
                transcript_id = record.feature

                if transcript_id not in vep_records:
                    vep_records[transcript_id] = []

                try:
                    record = record.convert_to_variant_record(anno, genome)
                except TranscriptionStopSiteMutationError:
                    continue
                except TranscriptionStartSiteMutationError:
                    continue
                except MNVParsingError:
                    warning(
                        f"MNVs are not currently supported. Skipping record: {record}"
                    )
                    continue

                vep_records[transcript_id].append(record)

        logger.info('VEP file %s loaded.', vep_file)

    if not vep_records:
        logger.warning('No variant record is saved.')
        return

    for records in vep_records.values():
        records.sort()

    logger.info('VEP sorting done.')

    metadata = common.generate_metadata(args)

    tx_rank = anno.get_transcript_rank()
    ordered_keys = sorted(vep_records.keys(), key=lambda x:tx_rank[x])

    all_records = []
    for key in ordered_keys:
        all_records.extend(vep_records[key])

    seqvar.io.write(all_records, output_path, metadata)

    logger.info('Variant info written to disk.')



""" Module for variant record with coordinate. """
from __future__ import annotations
from typing import TYPE_CHECKING
import math
from moPepGen.SeqFeature import FeatureLocation


if TYPE_CHECKING:
    from moPepGen.seqvar import VariantRecord

class VariantRecordWithCoordinate():
    """ This class models the variant record with its coordinate location at
    a gene or protein. This is used mainly in the graph to keep track on the
    location of variants of a node when the variable bubble expands forward
    or backward. """
    def __init__(self, variant:VariantRecord, location:FeatureLocation,
            is_stop_altering:bool=False, is_silent:bool=False,
            downstream_cleavage_altering:bool=False,
            upstream_cleavage_altering:bool=False):
        """ Constructor """
        self.variant = variant
        self.location = location
        self.is_stop_altering = is_stop_altering
        self.is_silent = is_silent
        self.downstream_cleavage_altering = downstream_cleavage_altering
        self.upstream_cleavage_altering = upstream_cleavage_altering

    def not_cleavage_altering(self) -> bool:
        """ Whether is not downstream or upstream cleavage altering """
        return not (self.upstream_cleavage_altering or self.downstream_cleavage_altering)

    def shift(self, index:int) -> VariantRecordWithCoordinate:
        """ Shift the coordinate of the object by a given number. """
        return self.__class__(
            variant=self.variant,
            location=FeatureLocation(
                start=max(self.location.start + index, 0),
                end=self.location.end + index,
                seqname=self.location.seqname,
                reading_frame_index=self.location.reading_frame_index,
                start_offset=self.location.start_offset,
                end_offset=self.location.end_offset
            ),
            is_stop_altering=self.is_stop_altering,
            is_silent=self.is_silent,
            upstream_cleavage_altering=self.upstream_cleavage_altering,
            downstream_cleavage_altering=self.downstream_cleavage_altering
        )

    def to_protein_coordinates(self) -> VariantRecordWithCoordinate:
        """ Returns a new object with the coordinates of the translated protein
        """
        start = math.floor(self.location.start / 3)
        end = math.ceil(self.location.end / 3)
        start_offset = self.location.start - start * 3
        end_offset = end * 3 - self.location.end
        return VariantRecordWithCoordinate(
            variant=self.variant,
            location=FeatureLocation(
                start=start, end=end, seqname=self.location.seqname,
                reading_frame_index=self.location.reading_frame_index,
                start_offset=start_offset, end_offset=end_offset
            ),
            is_stop_altering=self.is_stop_altering,
            is_silent=self.is_silent
        )

    def __getitem__(self, index) -> VariantRecordWithCoordinate:
        """ Get item """
        start, stop, _ = index.indices(self.location.end)
        start_index = max([start, self.location.start])
        end_index = min([stop, self.location.end])
        start_offset = self.location.start_offset if start == self.location.start else 0
        end_offset = self.location.end_offset if stop == self.location.end else 0
        location = FeatureLocation(
            start=start_index, end=end_index, seqname=self.location.seqname,
            reading_frame_index=self.location.reading_frame_index,
            start_offset=start_offset, end_offset=end_offset
        )
        return VariantRecordWithCoordinate(
            variant=self.variant,
            location=location,
            is_stop_altering=self.is_stop_altering,
            is_silent=self.is_silent
        )



""" `callNovelORF` calls noncanonical peptide sequences from novel ORFs.
It finds all start codons of any novel ORF gene. """
from __future__ import annotations
import argparse
from typing import TYPE_CHECKING, Set, List, Tuple, IO, Dict
from pathlib import Path
from Bio.SeqIO import FastaIO
from moPepGen import params, svgraph, aa, get_logger, VARIANT_PEPTIDE_SOURCE_DELIMITER
from moPepGen.dna.DNASeqRecord import DNASeqRecordWithCoordinates
from moPepGen.err import ReferenceSeqnameNotFoundError
from moPepGen.cli import common


if TYPE_CHECKING:
    from Bio.Seq import Seq
    from moPepGen.gtf import TranscriptAnnotationModel
    from moPepGen.dna import DNASeqDict

OUTPUT_FILE_FORMATS = ['.fa', '.fasta']

# pylint: disable=W0212
def add_subparser_call_novel_orf(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen callNovelORF """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='callNovelORF',
        help='Call non-canonical peptides from novel ORFs.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    p.add_argument(
        '-o', '--output-path',
        type=Path,
        help='Output path to the novel ORF peptide FASTA.'
        f" Valid formats: {OUTPUT_FILE_FORMATS}",
        metavar='<file>',
        required=True
    )
    p.add_argument(
        '--output-orf',
        type=Path,
        help='Output path to the FASTA file with novel ORF sequences.'
        f" Valid formats: {OUTPUT_FILE_FORMATS}",
        metavar='<file>',
        required=False,
        default=None
    )
    p.add_argument(
        '--min-tx-length',
        type=int,
        help='Minimal transcript length.',
        metavar='<number>',
        default=21
    )
    p.add_argument(
        '--orf-assignment',
        type=str,
        help='Defines how ORF assignment should be done. The last ORF upstream'
        ' to the peptide is used for `max` and the first (most upstream) one is'
        ' used for `min`',
        choices=['max', 'min'],
        default='max',
        metavar='<choice>'
    )
    p.add_argument(
        '--coding-novel-orf',
        action='store_true',
        help='Include coding transcripts to find alternative ORFs.'
    )
    p.add_argument(
        '--w2f-reassignment',
        action='store_true',
        help='Include peptides with W > F (Tryptophan to Phenylalanine) '
        'reassignment.'
    )
    p.add_argument(
        '--inclusion-biotypes',
        type=Path,
        help='Inclusion biotype list.',
        metavar='<file>',
        default=None
    )
    p.add_argument(
        '--exclusion-biotypes',
        type=Path,
        help='Exclusion biotype list.',
        metavar='<file>',
        default=None
    )

    common.add_args_reference(p)
    common.add_args_cleavage(p)
    common.add_args_debug_level(p)

    p.set_defaults(func=call_novel_orf_peptide)
    common.print_help_if_missing_args(p)
    return p

def call_novel_orf_peptide(args:argparse.Namespace) -> None:
    """ Main entrypoint for calling novel ORF peptide """
    logger = get_logger()
    common.validate_file_format(
        args.output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )
    if args.output_orf:
        common.validate_file_format(
            args.output_orf, OUTPUT_FILE_FORMATS, check_writable=True
        )

    cleavage_params = params.CleavageParams(
        enzyme=args.cleavage_rule,
        exception=args.cleavage_exception,
        miscleavage=int(args.miscleavage),
        min_mw=float(args.min_mw),
        min_length=args.min_length,
        max_length=args.max_length
    )

    common.print_start_message(args)

    genome, anno, proteome, canonical_peptides = common.load_references(
        args=args, load_proteome=True, cleavage_params=cleavage_params
    )

    inclusion_biotypes, exclusion_biotypes = common.load_inclusion_exclusion_biotypes(args)

    novel_orf_peptide_pool = aa.VariantPeptidePool()
    orf_pool = []

    i = 0
    for tx_id in anno.transcripts:
        tx_model = anno.transcripts[tx_id]
        if tx_model.is_protein_coding:
            if not args.coding_novel_orf:
                pass
        else:
            if inclusion_biotypes and \
                    tx_model.transcript.biotype not in inclusion_biotypes:
                continue
            if exclusion_biotypes and \
                    tx_model.transcript.biotype in exclusion_biotypes:
                continue
            if tx_id in proteome:
                continue
            if tx_model.transcript_len() < args.min_tx_length:
                continue

        try:
            peptides, orfs = call_noncoding_peptide_main(
                tx_id=tx_id, tx_model=tx_model, genome=genome,
                canonical_peptides=canonical_peptides,
                cleavage_params=cleavage_params,
                orf_assignment=args.orf_assignment,
                w2f_reassignment=args.w2f_reassignment
            )

            if not orfs:
                continue

            orf_pool.extend(orfs)

            for peptide in peptides:
                novel_orf_peptide_pool.add_peptide(peptide, canonical_peptides,
                    cleavage_params)
        except ReferenceSeqnameNotFoundError as e:
            if not ReferenceSeqnameNotFoundError.raised:
                logger.warning('%s: Make sure your GTF and FASTA files match.', e.args[0])
                ReferenceSeqnameNotFoundError.mute()
        except:
            logger.error('Exception raised from %s', tx_id)
            raise

        i += 1
        if i % 5000 == 0:
            logger.info('%i transcripts processed.', i)

    novel_orf_peptide_pool.write(args.output_path)
    if args.output_orf:
        with open(args.output_orf, 'w') as handle:
            write_orf(orf_pool, handle)

    logger.info('Noncanonical peptide FASTA file written to disk.')


def call_noncoding_peptide_main(tx_id:str, tx_model:TranscriptAnnotationModel,
        genome:DNASeqDict, canonical_peptides:Set[str],
        cleavage_params:params.CleavageParams, orf_assignment:str,
        w2f_reassignment:bool
        ) -> Tuple[Set[aa.AminoAcidSeqRecord],List[aa.AminoAcidSeqRecord]]:
    """ Call novel ORF peptides """
    chrom = tx_model.transcript.location.seqname
    try:
        contig_seq = genome[chrom]
    except KeyError as e:
        raise ReferenceSeqnameNotFoundError(chrom) from e

    tx_seq = tx_model.get_transcript_sequence(contig_seq)

    dgraph = svgraph.ThreeFrameTVG(
        seq=tx_seq,
        _id=tx_id,
        cds_start_nf=True,
        has_known_orf=False,
        cleavage_params=cleavage_params,
        gene_id=tx_model.gene_id,
        coordinate_feature_type='transcript',
        coordinate_feature_id=tx_id
    )
    dgraph.init_three_frames()
    pgraph = dgraph.translate()
    pgraph.create_cleavage_graph()
    peptide_anno = pgraph.call_variant_peptides(
        check_variants=False,
        check_orf=True,
        denylist=canonical_peptides,
        orf_assignment=orf_assignment,
        w2f=w2f_reassignment,
        check_external_variants=False
    )
    peptide_map:Dict[Seq, Set[str]] = {}
    for seq, annotated_labels in peptide_anno.items():
        for label in annotated_labels:
            if seq in peptide_map:
                peptide_map[seq].add(label.label)
            else:
                peptide_map[seq] = {label.label}
    peptides = set()
    for seq, labels in peptide_map.items():
        label = VARIANT_PEPTIDE_SOURCE_DELIMITER.join(labels)
        peptides.add(
            aa.AminoAcidSeqRecord(
                seq=seq,
                description=label,
                name=label
            )
        )
    orfs = get_orf_sequences(
        pgraph=pgraph,
        tx_id=tx_id,
        gene_id=tx_model.gene_id,
        tx_seq=tx_seq,
        exclude_canonical_orf=False
    )
    return peptides, orfs

def get_orf_sequences(pgraph:svgraph.PeptideVariantGraph, tx_id:str, gene_id:str,
        tx_seq:DNASeqRecordWithCoordinates, exclude_canonical_orf:bool
        ) -> List[aa.AminoAcidSeqRecord]:
    """ Get the full ORF sequences """
    seqs = []
    translate_seqs = [tx_seq[i:].translate() for i in range(3)]
    for orf, orf_id in pgraph.orf_id_map.items():
        orf_start = orf[0]
        if exclude_canonical_orf and tx_seq.orf and tx_seq.orf.start == orf_start:
            continue
        seq_start = int(orf_start / 3)
        reading_frame_index = orf_start % 3
        #seq_start = int((orf_start - node.orf[0]) / 3)
        translate_seq = translate_seqs[reading_frame_index]
        seq_len = translate_seq.seq[seq_start:].find('*')
        if seq_len == -1:
            seq_len = len(translate_seq.seq) - seq_start
        orf_end = orf_start + seq_len * 3
        seq_end = seq_start + seq_len
        seqname = f"{tx_id}|{gene_id}|{orf_id}|{orf_start}-{orf_end}"
        seq = translate_seq[seq_start:seq_end]
        seq.id = seqname
        seq.name = seqname
        seq.description = seqname
        seqs.append(seq)
    return seqs

def write_orf(orfs:List[aa.AminoAcidSeqRecord], handle:IO):
    """ Write ORF sequences """
    record2title = lambda x: x.description
    writer = FastaIO.FastaWriter(handle, record2title=record2title)
    for record in orfs:
        writer.write_record(record)



""" Validate callNovelORF results with bruteForceNoncoding """
import argparse
from contextlib import redirect_stdout
from datetime import datetime
from pathlib import Path
import random
from typing import IO, List, Set
from Bio import SeqIO
from moPepGen import get_logger
from moPepGen.cli import call_novel_orf_peptide, common
from moPepGen.gtf import GtfIO
from moPepGen.gtf.GTFSeqFeature import GTFSeqFeature
from moPepGen.util.common import load_references
from moPepGen.util.validate_variant_calling import call_downsample_reference
from moPepGen.util import brute_force_novel_orf


# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ parse args """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='validateNovelORFCalling',
        help='Validate the novel ORF peptide calling result of the graph-based'
        ' algorithm with the brute force algorithm.'
    )
    parser.add_argument(
        '--tx-id',
        type=str,
        nargs='*',
        metavar='<value>',
        help='Transcript IDs to sample from to validate. If not given, all'
        ' noncoding transcripts from the annotation GTF will be used.'
    )
    parser.add_argument(
        '--n-iter',
        type=int,
        help='Number of iterations',
        metavar='<number>',
        default=1
    )
    parser.add_argument(
        '-o', '--output-dir',
        type=Path,
        help='Path to the output diractory.',
        metavar='<file>'
    )
    common.add_args_reference(parser, proteome=True, index=False)
    parser.set_defaults(func=main)
    common.print_help_if_missing_args(parser)
    return parser

def get_transcript_ids(handle:IO) -> Set[str]:
    """ Get all transcript IDs. """
    _, exclusion_biotypes = common.load_inclusion_exclusion_biotypes(
        argparse.Namespace(inclusion_biotypes=None, exclusion_biotypes=None)
    )
    tx_ids:Set[str]= set()
    record:GTFSeqFeature
    for record in GtfIO.parse(handle):
        record.source = 'GENCODE'
        if record.type.lower() == 'gene':
            continue
        if record.biotype in exclusion_biotypes:
            continue
        tx_ids.add(record.transcript_id)
    return tx_ids

def call_novel_orf(ref_dir:Path, output_fasta:Path):
    """ call callNovelORF """
    args = argparse.Namespace()
    args.command = 'callNovelORF'
    args.genome_fasta = ref_dir/'genome.fasta'
    args.annotation_gtf = ref_dir/'annotation.gtf'
    args.proteome_fasta = ref_dir/'proteome.fasta'
    args.index_dir = None
    args.min_tx_length = 21
    args.cleavage_rule = 'trypsin'
    args.miscleavage = 2
    args.min_mw = 500.
    args.min_length = 7
    args.max_length = 25
    args.output_orf = None
    args.output_path = output_fasta
    args.min_tx_length = 21
    args.coding_novel_orf = False
    args.inclusion_biotypes = None
    args.exclusion_biotypes = None
    args.quiet = True
    call_novel_orf_peptide(args)

def call_brute_force_noncoding(tx_id:str, ref_dir:Path, output_path:Path):
    """ call bruteForceNoncoding """
    args = argparse.Namespace()
    args.tx_id = tx_id
    args.reference_dir = ref_dir
    args.canonical_peptides = None
    args.cleavage_rule = 'trypsin'
    args.miscleavage = 2
    args.min_mw = 500.
    args.min_length = 7
    args.max_length = 25

    with open(output_path, 'wt') as handle:
        with redirect_stdout(handle):
            brute_force_novel_orf.main(args)

def get_transcript_length(tx_id:str, ref_dir:Path) -> int:
    """ Get the transcript length """
    anno, genome, *_ = load_references(
        path_anno=ref_dir/'annotation.gtf',
        path_genome=ref_dir/'genome.fasta',
        path_proteome=ref_dir/'proteome.fasta'
    )
    return len(anno.transcripts[tx_id].get_transcript_sequence(genome['chr1']))

def assert_equal(noncoding_fasta:Path, brute_force_txt:Path, output_dir:Path) -> bool:
    """ Compare the results of callNovelORF and bruteForceNovelORF """
    with open(noncoding_fasta, 'rt') as handle:
        noncoding_seqs = set()
        for seq in SeqIO.parse(handle, 'fasta'):
            noncoding_seqs.add(str(seq.seq))

    with open(brute_force_txt, 'rt') as handle:
        brute_force_seqs = set()
        for line in handle:
            brute_force_seqs.add(line.rstrip())

    if noncoding_seqs != brute_force_seqs:
        noncoding_only = noncoding_seqs - brute_force_seqs
        brute_force_only = brute_force_seqs - noncoding_seqs

        if noncoding_only:
            noncoding_only_file = output_dir/'call_noncoding_only.txt'
            with open(noncoding_only_file, 'wt') as handle:
                for seq in noncoding_only:
                    handle.write(seq + '\n')

        if brute_force_only:
            brute_force_only_file = output_dir/'brute_force_only.txt'
            with open(brute_force_only_file, 'wt') as handle:
                for seq in brute_force_only:
                    handle.write(seq + '\n')

        return False
    return True

class ValidationRecord():
    """ Validation summary record """
    def __init__(self, tx_id:str=None, status:str=None, submitted:datetime=None,
            completed:datetime=None, tx_len:int=None,
            call_noncoding_start:datetime=None, call_noncoding_end:datetime=None,
            brute_force_start:datetime=None, brute_force_end:datetime=None):
        """ """
        self.tx_id = tx_id
        self.status = status
        self.submitted = submitted
        self.completed = completed
        self.tx_len = tx_len
        self.call_noncoding_start = call_noncoding_start
        self.call_noncoding_end = call_noncoding_end
        self.brute_force_start = brute_force_start
        self.brute_force_end = brute_force_end

    def submit(self):
        """ Validation starts """
        self.submitted = datetime.now()

    def complete(self, status:str):
        """ set the complete timestamp """
        self.completed = datetime.now()
        self.status = status

    def start_call_novel_orf(self):
        """ set timestamp when callNovelORF starts """
        self.call_noncoding_start = datetime.now()

    def end_call_novel_orf(self):
        """ set timestamp when callNovelORF ends """
        self.call_noncoding_end = datetime.now()

    def start_brute_force(self):
        """ set timestamp when bruteForce starts """
        self.brute_force_start = datetime.now()

    def end_brute_force(self):
        """ set timestamp when bruteForce ends """
        self.brute_force_end = datetime.now()

    def to_tsv(self) -> str:
        """ convert to TSV record """
        submitted = self.submitted.strftime('%Y-%m-%d %H:%M:%S') \
            if self.submitted else '-'
        completed = self.completed.strftime('%Y-%m-%d %H:%M:%S') \
            if self.completed else '-'
        call_noncoding_start = self.call_noncoding_start.strftime('%Y-%m-%d %H:%M:%S') \
            if self.call_noncoding_start else '-'
        call_noncoding_end = self.call_noncoding_end.strftime('%Y-%m-%d %H:%M:%S') \
            if self.call_noncoding_end else '-'
        call_noncoding_time = str(self.call_noncoding_end - self.call_noncoding_start) \
            if self.call_noncoding_start and self.call_noncoding_end else '-'
        brute_force_start = self.brute_force_start.strftime('%Y-%m-%d %H:%M:%S') \
            if self.brute_force_start else '-'
        brute_force_end = self.brute_force_end.strftime('%Y-%m-%d %H:%M:%S') \
            if self.brute_force_end else '-'
        brute_force_time = str(self.brute_force_end - self.brute_force_start) \
            if self.brute_force_start and self.brute_force_end else '-'
        return '\t'.join([
            self.tx_id, str(self.tx_len), self.status, submitted, completed,
            call_noncoding_start, call_noncoding_end, call_noncoding_time,
            brute_force_start, brute_force_end, brute_force_time
        ])


class ValidationSummary():
    """ Valiadtion summary """
    def __init__(self, data:List[ValidationRecord]=None):
        """ constructor """
        self.data = data or []

    def append(self, x:ValidationRecord):
        """ Add summary record """
        self.data.append(x)

    def write(self, path:Path):
        """ Write summary """
        header = '\t'.join([
            'tx_id', 'tx_len', 'status',
            'submitted', 'completed',
            'call_noncoding_start', 'call_noncoding_end', 'call_noncoding_time',
            'brute_force_start', 'brute_force_end', 'brute_force_time'
        ])
        with open(path, 'w') as handle:
            handle.write(header + '\n')
            for record in self.data:
                handle.write(record.to_tsv() + '\n')

def main(args:argparse.Namespace):
    """ main entrypoint """
    logger = get_logger()
    if args.tx_id:
        tx_ids = args.tx_id
    else:
        with open(args.annotation_gtf, 'rt') as handle:
            tx_ids = get_transcript_ids(handle)

    n_iter = min(args.n_iter, len(tx_ids))
    tx_ids_sampled = random.sample(tx_ids, n_iter)

    summary = ValidationSummary()

    for tx_id in tx_ids_sampled:
        output_dir:Path = args.output_dir/tx_id
        output_dir.mkdir(exist_ok=True)
        ref_dir = output_dir/'index'
        ref_dir.mkdir(exist_ok=True)

        record = ValidationRecord(tx_id=tx_id)
        summary.append(record)
        record.submit()

        noncoding_fasta = output_dir/'call_noncoding.fasta'
        brute_force_txt = output_dir/'brute_force_noncoding.fasta'

        call_downsample_reference(
            genome=args.genome_fasta,
            anno=args.annotation_gtf,
            protein=args.proteome_fasta,
            tx_id=tx_id,
            output_dir=ref_dir
        )

        record.tx_len = get_transcript_length(tx_id, ref_dir)

        record.start_call_novel_orf()
        call_novel_orf(
            ref_dir=ref_dir,
            output_fasta=noncoding_fasta
        )
        record.end_call_novel_orf()

        record.start_brute_force()
        call_brute_force_noncoding(
            tx_id=tx_id,
            ref_dir=ref_dir,
            output_path=brute_force_txt
        )
        record.end_brute_force()

        res = assert_equal(
            noncoding_fasta=noncoding_fasta,
            brute_force_txt=brute_force_txt,
            output_dir=output_dir
        )
        record.complete('SUCCEEDED' if res else 'FAILED')

        logger.info(
            "Transcript ID: %s, %s!",
            tx_id,
            'Equal' if res else 'Not equal'
        )

    summary.write(args.output_dir/'validate_noncoding_summary.tsv')



""" Module for command line interface """
from .generate_index import add_subparser_generate_index, generate_index
from .call_variant_peptide import add_subparser_call_variant, call_variant_peptide
from .call_novel_orf import add_subparser_call_novel_orf, call_novel_orf_peptide
from .call_alt_translation import add_subparser_call_alt_translation, call_alt_translation
from .parse_vep import add_subparser_parse_vep, parse_vep
from .parse_reditools import add_subparser_parse_reditools, parse_reditools
from .parse_star_fusion import add_subparser_parse_star_fusion, parse_star_fusion
from .parse_fusion_catcher import add_subparser_parse_fusion_catcher, parse_fusion_catcher
from .parse_arriba import add_subparser_parse_arriba, parse_arriba
from .parse_rmats import add_subparser_parse_rmats, parse_rmats
from .parse_circexplorer import add_subparser_parse_circexplorer, parse_circexplorer
from .split_fasta import add_subparser_split_fasta, split_fasta
from .filter_fasta import add_subparser_filter_fasta, filter_fasta
from .index_gvf import add_subparser_index_gvf, index_gvf
from .merge_fasta import add_subparser_merge_fasta, merge_fasta
from .encode_fasta import add_subparser_encode_fasta, encode_fasta
from .decoy_fasta import add_subparser_decoy_fasta, decoy_fasta
from .summarize_fasta import add_subparser_summarize_fasta, summarize_fasta
from .update_index import add_subparser_update_index, update_index



""" Brute force caller for noncanonical peptides of noncoding gene """
import argparse
import re
from typing import List
from pathlib import Path
import pickle
import sys
from moPepGen import aa, gtf, dna
from moPepGen.aa.AminoAcidSeqRecord import AminoAcidSeqRecord
from moPepGen.cli import common


# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ parse command line arguments """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='bruteForceNovelORF',
        help='Call noncanonical peptides from noncoding gnees with the brute'
        'force algorithm'
    )
    parser.add_argument(
        '-t', '--tx-id',
        type=str,
        help='Transcript ID'
    )
    parser.add_argument(
        '-r', '--reference-dir',
        type=Path,
        help='Reference directory. Must contain genome.fa, annotation.gtf, and'
        ' proteome.fasta. The directory should be generated by'
        ' the downsampleReference command'
    )
    parser.add_argument(
        '--canonical-peptides',
        type=Path,
        help='Path to the pickled canonical peptide sequence.',
        default=None
    )
    common.add_args_cleavage(parser)
    common.add_args_debug_level(parser)
    parser.set_defaults(func=main)
    common.print_help_if_missing_args(parser)
    return parser

def main(args:argparse.Namespace):
    """ main entrance """
    anno = gtf.GenomicAnnotation()
    anno.dump_gtf(args.reference_dir/'annotation.gtf')

    genome = dna.DNASeqDict()
    genome.dump_fasta(args.reference_dir/'genome.fasta')

    proteome = aa.AminoAcidSeqDict()
    proteome.dump_fasta(args.reference_dir/'proteome.fasta')

    rule = args.cleavage_rule
    exception = args.cleavage_exception

    if args.canonical_peptides:
        with open(args.canonical_peptides, 'rb') as handle:
            canonical_pool = pickle.load(handle)
    else:
        canonical_pool = None

    tx_id = args.tx_id
    tx_model = anno.transcripts[tx_id]
    tx_seq = tx_model.get_transcript_sequence(genome[tx_model.transcript.chrom])

    noncanonical_peptides = set()

    three_frame_translates:List[AminoAcidSeqRecord] = []
    for i in range(3):
        j = len(tx_seq.seq) - (len(tx_seq.seq) - i) % 3
        three_frame_translates.append(tx_seq[i:j].translate(to_stop=False))

    for frame_seq in three_frame_translates:
        for it in re.finditer('M', str(frame_seq.seq)):
            i = it.start()
            j = frame_seq.seq[i:].find('*')
            if j == -1:
                j = len(frame_seq.seq)
            else:
                j += i
            aa_seq = frame_seq[i:j]
            peptides = aa_seq.enzymatic_cleave(rule, exception)
            for peptide in peptides:
                if peptide is peptides[0] and peptide.seq.startswith('M'):
                    seq = str(peptide.seq[1:])
                    if canonical_pool and seq not in canonical_pool:
                        noncanonical_peptides.add(seq)
                seq = str(peptide.seq)
                if not canonical_pool or seq not in canonical_pool:
                    noncanonical_peptides.add(str(peptide.seq))

    noncanonical_peptides = sorted(noncanonical_peptides)
    for peptide in noncanonical_peptides:
        print(peptide, file=sys.stdout)



""" moPepGen constants """
from enum import Enum


class VariantPrefix(Enum):
    """ Varinat Prefix"""
    SNV    = 'SNV'
    INDEL  = 'INDEL'
    MNV    = 'MNV'
    RES    = 'RES'
    SE     = 'SE'
    RI     = 'RI'
    A3SS   = 'A3SS'
    A5SS   = 'A5SS'
    MXE    = 'MXE'
    W2F    = 'W2F'
    SECT   = 'SECT'
    CIRC   = 'CIRC'
    CI     = 'CI'
    FUSION = 'FUSION'

    def __str__(self):
        """ str """
        return self.value

    @classmethod
    def ctbv(cls):
        """ Conserved Transcript Backbone Variants """
        return [
            cls.SNV,
            cls.INDEL,
            cls.MNV,
            cls.RES,
            cls.SE,
            cls.RI,
            cls.A3SS,
            cls.A5SS,
            cls.MXE,
            cls.W2F,
            cls.SECT
        ]

    @classmethod
    def ntbv(cls):
        """ Novel Transcript Backbone Variants """
        return [
            cls.CIRC,
            cls.CI,
            cls.FUSION
        ]

    @classmethod
    def alt_translation(cls):
        """ Alternative translation variants """
        return [
            cls.W2F,
            cls.SECT
        ]

# Variant related constants
SINGLE_NUCLEOTIDE_SUBSTITUTION = ['SNV', 'SNP', 'INDEL', 'MNV', 'RNAEditingSite']
ATTRS_POSITION = ['START', 'DONOR_START', 'ACCEPTER_START', 'ACCEPTER_POSITION']
ALTERNATIVE_SPLICING_TYPES = ['Insertion', 'Deletion', 'Substitution']
RMATS_TYPES = ['SE', 'RI', 'A3SS', 'A5SS', 'MXE']
CODON_REASSIGNMENTS_TYPES = ['W2F']
SEC_TERMINATION_TYPE = 'SECT'

# Variant sources
SOURCE_NOVEL_ORF = 'NovelORF'
SOURCE_CODON_REASSIGNMENT = 'CodonReassign'
SOURCE_SEC_TERMINATION = 'SECT'

# CLI
PROG_NAME = 'moPepGen'



""" ExPASy's PeptideCutter rules. Adopted from pyteomics at:
https://github.com/levitsky/pyteomics """
from typing import Dict, Tuple


EXPASY_RULES: Dict[str, str] = {
    'arg-c': r'R',
    'asp-n': r'\w(?=D)',
    'bnps-skatole': r'W',
    'caspase 1': r'(?<=[FWYL]\w[HAT])D(?=[^PEDQKR])',
    'caspase 2': r'(?<=DVA)D(?=[^PEDQKR])',
    'caspase 3': r'(?<=DMQ)D(?=[^PEDQKR])',
    'caspase 4': r'(?<=LEV)D(?=[^PEDQKR])',
    'caspase 5': r'(?<=[LW]EH)D',
    'caspase 6': r'(?<=VE[HI])D(?=[^PEDQKR])',
    'caspase 7': r'(?<=DEV)D(?=[^PEDQKR])',
    'caspase 8': r'(?<=[IL]ET)D(?=[^PEDQKR])',
    'caspase 9': r'(?<=LEH)D',
    'caspase 10': r'(?<=IEA)D',
    'chymotrypsin high specificity': r'([FY](?=[^P]))|(W(?=[^MP]))',
    'chymotrypsin low specificity': r'([FLY](?=[^P]))|(W(?=[^MP]))|'
    r'(M(?=[^PY]))|(H(?=[^DMPW]))',
    'clostripain': r'R',
    'cnbr': r'M',
    'enterokinase': r'(?<=[DE]{3})K',
    'factor xa': r'(?<=[AFGILTVM][DE]G)R',
    'formic acid': r'D',
    'glutamyl endopeptidase': r'E',
    'granzyme b': r'(?<=IEP)D',
    'hydroxylamine': r'N(?=G)',
    'iodosobenzoic acid': r'W',
    'lysc': r'K',
    'lysn': r'\w(?=K)',
    'ntcb': r'\w(?=C)',
    'pepsin ph1.3':  r'((?<=[^HKR][^P])[^R](?=[FL][^P]))|'
    r'((?<=[^HKR][^P])[FL](?=\w[^P]))',
    'pepsin ph2.0': r'((?<=[^HKR][^P])[^R](?=[FLWY][^P]))|'
    r'((?<=[^HKR][^P])[FLWY](?=\w[^P]))',
    'proline endopeptidase': r'(?<=[HKR])P(?=[^P])',
    'proteinase k': r'[AEFILTVWY]',
    'staphylococcal peptidase i': r'(?<=[^E])E',
    'thermolysin': r'[^DE](?=[AFILMV])',
    'thrombin': r'((?<=G)R(?=G))|'
    r'((?<=[AFGILTVM][AFGILTVWA]P)R(?=[^DE][^DE]))',
    'trypsin': r'([KR](?=[^P]))|((?<=W)K(?=P))|((?<=M)R(?=P))',
    'trypsin_exception': r'((?<=[CD])K(?=D))|((?<=C)K(?=[HY]))|((?<=C)R(?=K))|'
    r'((?<=R)R(?=[HR]))',
}

EXPASY_RULES2: Dict[str, str] = {
    'arg-c': r'R',
    'asp-n': r'\wD',
    'bnps-skatole': r'W',
    'caspase 1': r'[FWYL]\w[HAT]D[^PEDQKR]',
    'caspase 2': r'DVAD[^PEDQKR]',
    'caspase 3': r'DMQD[^PEDQKR]',
    'caspase 4': r'LEVD[^PEDQKR]',
    'caspase 5': r'[LW]EHD',
    'caspase 6': r'VE[HI]D[^PEDQKR]',
    'caspase 7': r'DEVD[^PEDQKR]',
    'caspase 8': r'[IL]ETD[^PEDQKR]',
    'caspase 9': r'LEHD',
    'caspase 10': r'IEAD',
    'chymotrypsin high specificity': r'([FY][^P])|(W[^MP])',
    'chymotrypsin low specificity': r'([FLY][^P])|(W[^MP])|'
    r'(M[^PY])|(H[^DMPW])',
    'clostripain': r'R',
    'cnbr': r'M',
    'enterokinase': r'[DE]{3}K',
    'factor xa': r'[AFGILTVM][DE]GR',
    'formic acid': r'D',
    'glutamyl endopeptidase': r'E',
    'granzyme b': r'IEPD',
    'hydroxylamine': r'NG',
    'iodosobenzoic acid': r'W',
    'lysc': r'K',
    'lysn': r'\wK',
    'ntcb': r'\wC',
    'pepsin ph1.3':  r'([^HKR][^P][^R][FL][^P])|([^HKR][^P][FL]\w[^P])',
    'pepsin ph2.0': r'([^HKR][^P][^R][FLWY][^P])|([^HKR][^P][FLWY]\w[^P])',
    'proline endopeptidase': r'[HKR]P[^P]',
    'proteinase k': r'[AEFILTVWY]',
    'staphylococcal peptidase i': r'[^E]E',
    'thermolysin': r'[^DE][AFILMV]',
    'thrombin': r'(GRG)|([AFGILTVM][AFGILTVWA]PR[^DE][^DE])',
    'trypsin': r'([KR][^P])|(WKP)|(MRP)',
    'trypsin_exception': r'([CD]KD)|(CK[HY])|(CRK)|(RR[HR])',
}

EXPASY_RULES_WINGS_SIZE: Dict[str, Tuple[int, int]] = {
    'arg-c': (1, 0),
    'asp-n': (0, 1),
    'bnps-skatole': (1, 0),
    'caspase 1': (4, 1),
    'caspase 2': (2, 1),
    'caspase 3': (2, 1),
    'caspase 4': (2, 1),
    'caspase 5': (2, 0),
    'caspase 6': (3, 1),
    'caspase 7': (2, 1),
    'caspase 8': (4, 1),
    'caspase 9': (4, 0),
    'caspase 10': (4, 0),
    'chymotrypsin high specificity': (1, 1),
    'chymotrypsin low specificity': (1, 1),
    'clostripain': (1, 0),
    'cnbr': (1, 0 ),
    'enterokinase': (4, 0),
    'factor xa': (4, 0),
    'formic acid': (1, 0),
    'glutamyl endopeptidase': (1, 0),
    'granzyme b': (4, 0),
    'hydroxylamine': (1, 1),
    'iodosobenzoic acid': (1, 0),
    'lysc': (1, 0),
    'lysn': (1, 1),
    'ntcb': (0, 1),
    'pepsin ph1.3': (3, 2),
    'pepsin ph2.0': (3, 2),
    'proline endopeptidase': (2, 1),
    'proteinase k': (1, 0),
    'staphylococcal peptidase i': (2, 0),
    'thermolysin': (1, 1),
    'thrombin': (4, 2),
    'trypsin': (2, 1),
    'trypsin_exception': (2, 1),
}
